#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""File inspection CLI & GUI."""

import click

import mesoscopy.inspect.data_viewers as dvs
from mesoscopy import io


@click.command("inspect")
@click.argument(
    "input_path",
    type=click.Path(exists=True),
)
@click.option(
    "--meta-only",
    "info_level",
    flag_value="meta_only",
    default=True,
)
@click.option(
    "--acquisition",
    "info_level",
    flag_value="acquisition",
)
@click.option(
    "--deltaf",
    "info_level",
    flag_value="deltaf",
)
@click.option(
    "--registered",
    "info_level",
    flag_value="registered",
)
def inspect_cmd(input_path: str, info_level: str) -> None:
    """Inspect a calcium recording session and associated preprocessing output. Files must be in NWB format."""
    inspect(input_path=input_path, info_level=info_level)


def inspect(input_path: str, info_level: str = "meta_only") -> None:  # noqa: PLR0911
    """Inspect a mesoscale recording NWB file, alongside any associated preprocessing data, using the inspection GUI.

    Args:
        input_path (str): Path to NWB file.
        info_level (str): Inspection level. Defaults to metadata only.
    """
    nwbfile = io.read_nwb(input_path)

    click.echo("-" * (len(input_path) + 14))
    click.secho(f"Metadata for {input_path}", bold=True)
    click.echo("-" * (len(input_path) + 14))
    click.echo("")
    click.echo(f"File identifier: \t {nwbfile.identifier}")
    click.echo(f"Session date: \t\t {nwbfile.session_start_time.date()}")
    click.echo(f"Session description: \t {nwbfile.session_description}")
    click.echo(
        f"Experimenter: \t\t {nwbfile.experimenter[0] if nwbfile.experimenter else 'Unknown'} ({nwbfile.lab}, {
            nwbfile.institution
        })"
    )
    click.echo("")
    click.echo(f"Subject ID: \t\t {nwbfile.subject.subject_id if nwbfile.subject else 'Unknown'}")
    click.echo(f"Subject DOB: \t\t {nwbfile.subject.date_of_birth.date() if nwbfile.subject else 'Unknown'}")
    click.echo(f"Subject sex: \t\t {nwbfile.subject.sex if nwbfile.subject else 'Unknown'}")
    click.echo(f"Subject species: \t {nwbfile.subject.species if nwbfile.subject else 'Unknown'}")
    click.echo(f"Subject strain: \t {nwbfile.subject.strain if nwbfile.subject else 'Unknown'}")
    click.echo(f"Subject genotype: \t {nwbfile.subject.genotype if nwbfile.subject else 'Unknown'}")
    click.echo("")
    click.echo(f"Behavior trials: \t {f'{len(nwbfile.trials.id)} trials' if nwbfile.trials else 'None found'}")
    click.echo(
        f"Acquisition modules: \t {', '.join(list(nwbfile.acquisition.keys())) if nwbfile.acquisition else 'None found'}"
    )
    click.echo(
        f"Processing modules: \t {', '.join(list(nwbfile.processing.keys())) if nwbfile.processing else 'None found'}"
    )
    click.echo(f"Analysis modules: \t {', '.join(list(nwbfile.analysis.keys())) if nwbfile.analysis else 'None found'}")
    click.echo("")
    if nwbfile.processing and nwbfile.processing.get("ophys"):
        click.echo(click.style("Mesoscopy processing progress:", bold=True))
        click.echo(
            f"\t{
                f'{click.style("✅ Preprocessed", fg="green")}'
                if nwbfile.processing.get('ophys').get('DeltaFSeries')
                else f'{click.style("❌ Not preprocessed", fg="red")}'
            }"
        )
        click.echo(
            f"\t{
                f'{click.style("✅ Registered", fg="green")}'
                if 'CCFRegisteredSeries' in nwbfile.processing.get('ophys').data_interfaces
                else f'{click.style("❌ Not registered", fg="red")}'
            }"
        )
        click.echo(
            f"\t{
                f'{click.style(f"✅ Analysed ({', '.join(list(nwbfile.analysis.keys()))})", fg="green")}'
                if nwbfile.analysis
                else f'{click.style("❌ No analysis found", fg="red")}'
            }"
        )
    else:
        click.echo(click.style("No mesoscopy processing found.", bold=True))

    click.echo("")

    if info_level == "meta_only":
        return None
    if info_level == "acquisition":
        if not nwbfile.acquisition.get("DualChannelImagingSeries"):
            return click.secho("⚠️  No acquisition data found!", fg="red", bold=True)
        click.echo("Launching acquisition viewer...")
        return dvs.acquisition_viewer(nwbfile)
    if info_level == "deltaf":
        if not nwbfile.processing.get("ophys").get("DeltaFSeries"):
            return click.secho("⚠️  No DeltaFSeries found!", fg="red", bold=True)
        click.echo("Launching ∆F viewer...")
        return dvs.deltaf_viewer(nwbfile)
    if info_level == "registered":
        if not nwbfile.processing.get("ophys").get("CCFRegisteredSeries"):
            return click.secho("⚠️  No CCFRegisteredSeries found!", fg="red", bold=True)
        click.echo("Launching CCF registered ∆F viewer...")
        return dvs.ccfregistered_viewer(nwbfile)
    return None



from importlib import resources

import numpy as np
from click.testing import CliRunner

import mesoscopy
import mesoscopy.register as reg
import mesoscopy.resources as res


def test_landmarks_affine():
    """Test affine transform calculation for defined landmarks"""


def test_load_preprocessed_h5(preproc_h5):
    session_id, deltaf, ts = reg.load_deltaf(preproc_h5)
    assert session_id == "preproc"
    assert deltaf.shape == (300, 40, 40)
    assert ts.shape == (300,)


def test_load_preprocessed_nwb(preproc_nwb):
    session_id, deltaf, ts = reg.load_deltaf(preproc_nwb, nwb=True)
    assert session_id == "session_1234"
    assert deltaf.shape == (300, 40, 40)
    assert ts.shape == (300,)


def test_update_nwb(nwbfile, preproc_h5):
    tform_mock = np.array([[(1, 2, 3), (1, 2, 4)]])
    nwb = reg.update_nwb(nwbfile, preproc_h5, tform_mock)
    assert nwb.processing["ophys"]["CCFRegisteredSeries"].corrected.data
    assert nwb.processing["ophys"]["CCFRegisteredSeries"].original.data
    assert nwb.processing["ophys"]["CCFRegisteredSeries"].xy_translation.data.any()


def test_register_landmarks_cli_default_points_h5(preproc_h5, output_dir):
    mock_recording_landmarks = str(resources.files(res).joinpath("ccf_template_landmarks_140x142.csv"))
    runner = CliRunner()
    result = runner.invoke(
        mesoscopy.cli,
        args=f"register landmarks {preproc_h5} -r {mock_recording_landmarks} -o {output_dir}",
    )
    assert result.exit_code == 0


def test_mark_landmarks_gui(): ...



#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""HDF5 conversion utilities."""

import os
from datetime import datetime

import click
from pynwb import NWBFile
from pynwb.file import Subject
from pynwb.ophys import OnePhotonSeries
from pynwb.ophys import OpticalChannel

import mesoscopy.convert.metadata as mtd
from mesoscopy import io


def to_nwb(
    input_path: str,
    out_dir: str,
    link_only: bool = False,
    frames_group: str = "frames",
    timestamps_group: str = "timestamps",
    meta_path: str | None = None,
    subject_id: str | None = None,
    sex: str | None = None,
    genotype: str | None = None,
    species: str | None = None,
    strain: str | None = None,
    dob: str | None = None,
    session_description: str | None = None,
    experimenter: str | None = None,
    lab: str | None = None,
    institution: str | None = None,
    **kwargs,
) -> str:
    """Convert a mesoscale recording session from HDF5 to NWB. Optionally add session metadata.

    Metadata can be parsed from a compatible YAML or JSON file. Individual metadata arguments passed to this function (e.g. subject ID) will take precedence over the contents of a metadata file, if both are provided.

    Args:
        input_path (str): Path to raw HDF5 file.
        out_dir (str): Output directory.
        link_only (bool): Create links to HDF5 datasets instead of copying them over. Defaults to False.
        frames_group (str): HDF group path under which frame data is stored. Defaults to "frames".
        timestamps_group (str): HDF group path under which timestamp data is stored. Defaults to "timestamps".
        meta_path (typing.Optional[str], optional): Path to metadata file in JSON or YAML format. Defaults to None.
        subject_id (typing.Optional[str], optional): Subject ID. Defaults to None.
        sex (typing.Optional[str], optional): Subject sex. Defaults to None.
        genotype (typing.Optional[str], optional): Subject genotype. Defaults to None.
        species (typing.Optional[str], optional): Subject species. Defaults to None.
        strain (typing.Optional[str], optional): Subject strain. Defaults to None.
        dob (typing.Optional[str], optional): Subject date of birth. Defaults to None.
        session_description (typing.Optional[str], optional): Session description. Defaults to None.
        experimenter (typing.Optional[str], optional): Experimenter name. Defaults to None.
        lab (typing.Optional[str], optional): Laboratory experiment was done in. Defaults to None.
        institution (typing.Optional[str], optional): Institution experiment was done in. Defaults to None

    Raises:
        ValueError: Raised if declared frames or timestamp groups don't exist in the HDF5 file.

    Returns:
        str: Path to new NWB file.
    """
    h5file = io.read_h5(input_path)

    # Validate frame and timestamp groups actually exist.
    if (frames_group not in h5file) or (timestamps_group not in h5file):
        raise ValueError("Could not find frame or timestamp data, did you use the right group names?")

    session_start_time = datetime.fromisoformat(
        h5file[f"/{timestamps_group}"][0].decode("utf-8")
    )  # Use first timestamp as session start

    session_identifier = input_path.split(os.sep)[-1].replace(".h5", "")
    subject_meta = mtd.DEFAULT_METADATA.copy()

    if meta_path:
        if meta_path.endswith(".yaml") or meta_path.endswith(".yml"):
            subject_meta = mtd.read_yaml(meta_path)
        elif meta_path.endswith(".json"):
            subject_meta = mtd.read_json(meta_path)
        else:
            click.echo("WARNING - Invalid metadata file provided, skipping...")

    # Infer subject_id if not provided.
    if not subject_id and not meta_path:
        try:
            # Check if NWB-style file-naming
            if "sub-" in input_path.split(os.sep)[-1]:
                subject_meta["subject_id"] = (
                    input_path.split(os.sep)[-1].split("_")[0].replace("sub-", "").replace(".h5", "")
                )
        except IndexError:
            click.echo(
                "WARNING - Subject ID not provided and could not be inferred, using a default placeholder value. This might get confusing!"
            )

    if subject_id:
        subject_meta["subject_id"] = subject_id
    if sex:
        subject_meta["sex"] = sex
    if genotype:
        subject_meta["genotype"] = genotype
    if species:
        subject_meta["species"] = species
    if strain:
        subject_meta["strain"] = strain
    if dob:
        subject_meta["dob"] = dob
    if session_description:
        subject_meta["session_description"] = session_description
    if experimenter:
        subject_meta["experimenter"] = experimenter
    if lab:
        subject_meta["lab"] = lab
    if institution:
        subject_meta["institution"] = institution

    nwbfile = NWBFile(
        session_description=subject_meta.get("session_description"),
        identifier=session_identifier,
        session_start_time=session_start_time,  # Use first timestamp as session start
        experimenter=subject_meta.get("experimenter"),
        lab=subject_meta.get("lab"),
        institution=subject_meta.get("institution"),
    )

    nwbfile.subject = Subject(
        subject_id=subject_meta.get("subject_id"),
        species=subject_meta.get("species"),
        strain=subject_meta.get("strain"),
        sex=subject_meta.get("sex"),
        date_of_birth=datetime.fromisoformat(subject_meta.get("dob", "1900-01-01")),
        genotype=subject_meta.get("genotype"),
    )

    device = nwbfile.create_device(
        name="Mesoscope",
        description="Single-photon widefield imaging scope.",
        manufacturer="",
    )

    optical_channel = OpticalChannel(
        name="DualAcquisitionChannel",
        description="Acquisition channel for GCaMP6s excited at 470 and 405 nm.",
        emission_lambda=500.0,
    )

    imaging_plane = nwbfile.create_imaging_plane(
        name="DualChannelImagingPlane",
        optical_channel=optical_channel,
        # imaging_rate=50.,
        description="Dual excitation through the skull at 470 and 405 nm (25Hz per LED).",
        device=device,
        excitation_lambda=470.0,
        indicator="GCaMP6s",
        location="dorsal cortex",
        grid_spacing=[20.0, 20.0],
        grid_spacing_unit="micrometers",
    )

    source_data = h5file[f"/{frames_group}"]
    source_timestamps = h5file[f"/{timestamps_group}"]

    # recalculate timestamps to relative from session start
    timestamps = [
        (datetime.fromisoformat(timestamp.decode("utf-8")) - session_start_time).total_seconds()
        for timestamp in source_timestamps
    ]

    imaging_series = OnePhotonSeries(
        name="DualChannelImagingSeries",
        imaging_plane=imaging_plane,
        data=source_data,
        timestamps=timestamps,
        unit="pixel_intensity",
        binning=2,
        power=10.0,
        exposure_time=0.01,
        pmt_gain=10.0,
        description="Widefield cortical images acquired at alternating 470 and 405 nm excitation.",
        comments="This imaging series requires channel separation and haemodynamic subtraction before use.",
    )

    nwbfile.add_acquisition(imaging_series)

    nwb_path = out_dir + os.sep + session_identifier + ".nwb"
    io.write_nwb(
        path=nwb_path,
        nwbfile=nwbfile,
        link_data=link_only,
    )

    return nwb_path



#  Copyright (c) 2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
"""Module for quality assurance (QA) functions in the mesoscopy registration pipeline."""

import numpy.typing as npt
import plotly.express as px
import plotly.graph_objects as go


def plot_landmarks(
    source_landmarks: npt.NDArray,
    target_landmarks: npt.NDArray,
    as_html: bool = False,
) -> str | go.Figure:
    fig = go.Figure(
        layout=go.Layout(
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
            yaxis={
                "autorange": "reversed",
            },
        )
    )

    fig.add_trace(
        go.Scatter(x=source_landmarks[:, 1], y=source_landmarks[:, 0], mode="markers", name="Source landmarks")
    )
    fig.add_trace(
        go.Scatter(x=target_landmarks[:, 1], y=target_landmarks[:, 0], mode="markers", name="Target landmarks")
    )

    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_frame(data: npt.NDArray, landmarks=npt.NDArray | None, as_html: bool = False) -> str | go.Figure:
    """Plots a single frame of data using Plotly.

    Args:
        data (npt.NDArray): The frame data to plot, as a NumPy array.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = px.imshow(data)
    fig.update_layout(
        {
            "margin": {"l": 20, "r": 20, "t": 20, "b": 20},
        }
    )

    if landmarks:
        fig.add_trace(
            go.Scatter(
                x=landmarks[:, 1],
                y=landmarks[:, 0],
                mode="markers",
                name="Template landmarks",
                marker={"color": "Green"},
            )
        )

    if as_html:
        return fig.to_html(full_html=False)
    return fig



import importlib
import importlib.resources
import json
from datetime import datetime
from datetime import timedelta

import h5py as h5
import numpy as np
import pytest
import yaml
from dateutil.tz import tzlocal
from pynwb import NWBHDF5IO
from pynwb import NWBFile
from pynwb.ophys import ImageSeries
from pynwb.ophys import OnePhotonSeries
from pynwb.ophys import OpticalChannel

import tests.resources


@pytest.fixture(scope="session")
def random_idx():
    """Return a random index for testing."""
    return np.sort(np.random.choice(600, 300, replace=False))


@pytest.fixture
def nwbfile(tmp_path_factory, random_idx):
    """Create an NWBFile object for testing."""
    # Create a temporary file
    tmpfile = tmp_path_factory.mktemp("data") / "test.nwb"
    # Create an NWBFile object
    session_start_time = datetime(2024, 1, 1, 14, 0, 0, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="Test file, not real data",
        identifier="session_1234",
        session_start_time=session_start_time,
        # session_id="session_1234",
    )

    # Add fake raw imaging data
    device = nwbfile.create_device(
        name="Mesoscope",
        description="Single-photon widefield imaging scope.",
        manufacturer="INSS UK",
    )

    optical_channel = OpticalChannel(
        name="DualAcquisitionChannel",
        description="Acquisition channel for GCaMP6s excited at 470 and 405 nm.",
        emission_lambda=500.0,
    )

    imaging_plane = nwbfile.create_imaging_plane(
        name="DualChannelImagingPlane",
        optical_channel=optical_channel,
        # imaging_rate=50.,
        description="Random data for testing.",
        device=device,
        excitation_lambda=470.0,
        indicator="GCaMP6s",
        location="dorsal cortex",
        grid_spacing=[20.0, 20.0],
        grid_spacing_unit="micrometers",
    )

    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 40, 40))
    mock_isosb = np.random.normal(65, 50, size=(300, 40, 40))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    frames_num = 600
    timestamps = [(timedelta(milliseconds=i)).total_seconds() for i in range(frames_num)]

    imaging_series = OnePhotonSeries(
        name="DualChannelImagingSeries",
        imaging_plane=imaging_plane,
        data=mock_dual_channel,
        timestamps=timestamps,
        unit="pixel_intensity",
        binning=2,
        power=10.0,
        exposure_time=0.01,
        pmt_gain=10.0,
        description="Random data for testing.",
        comments="Completely random, no std or mean differentiation between test channels.",
    )

    nwbfile.add_acquisition(imaging_series)

    # Write the NWBFile object to file
    with NWBHDF5IO(str(tmpfile), "w") as io:
        io.write(nwbfile)
    # Return the path to the temporary file
    return str(tmpfile)


@pytest.fixture
def nwbfile_noacquisition(tmp_path_factory):
    """Create an NWBFile object for testing."""
    # Create a temporary file
    tmpfile = tmp_path_factory.mktemp("data") / "test.nwb"
    # Create an NWBFile object
    session_start_time = datetime(2024, 1, 1, 14, 0, 0, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="Test file, not real data",
        identifier="session_1234",
        session_start_time=session_start_time,
        # session_id="session_1234",
    )

    # Write the NWBFile object to file
    with NWBHDF5IO(str(tmpfile), "w") as io:
        io.write(nwbfile)
    # Return the path to the temporary file
    return str(tmpfile)


@pytest.fixture(scope="session")
def raw_h5(tmp_path_factory, random_idx):
    """Create an HDF5 file with fake raw data for testing."""
    # Create a temporary file
    tmpfile = tmp_path_factory.mktemp("data") / "preproc_test.h5"
    # Create timestamps
    timestamps = [
        (datetime.now() + timedelta(milliseconds=i)).isoformat().replace("-", "").replace(":", "").replace(".", "")
        for i in range(600)
    ]
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 40, 40))
    mock_isosb = np.random.normal(65, 50, size=(300, 40, 40))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    # Create an HDF5 file
    with h5.File(str(tmpfile), "w") as f:
        f.create_dataset("frames", data=mock_dual_channel)
        f.create_dataset("timestamps", data=timestamps, dtype="S26")
        f.create_dataset("test_isosb_idx", data=random_idx)

    # Return the path to the temporary file
    return str(tmpfile)


@pytest.fixture
def raw_avi():
    with importlib.resources.path(tests.resources, "example_recording.avi") as path:
        return str(path)


@pytest.fixture
def raw_timestamps():
    with importlib.resources.path(tests.resources, "example_recording_ts.csv") as path:
        return str(path)


@pytest.fixture
def preproc_h5(tmp_path_factory):
    """Create an HDF5 file with fake preprocessed data for testing."""
    # Create a temporary file
    tmpfile = tmp_path_factory.mktemp("data") / "preproc.h5"
    frames_num = 300
    timestamps = [(timedelta(milliseconds=i)).total_seconds() for i in range(frames_num)]
    # Create an HDF5 file
    with h5.File(str(tmpfile), "w") as f:
        f.create_dataset("/F", data=np.random.rand(frames_num, 40, 40))
        f.create_dataset("/timestamps", data=timestamps)
    # Return the path to the temporary file
    return str(tmpfile)


@pytest.fixture
def preproc_nwb(nwbfile, preproc_h5):
    f = h5.File(preproc_h5, "r")
    io = NWBHDF5IO(nwbfile, "a")
    nwb = io.read()
    deltaF_series = ImageSeries(
        name="DeltaFSeries",
        data=f["/F"][:],
        timestamps=f["/timestamps"][:],
        unit="df/f",
        description="dF/F widefield cortical imaging series.",
        comments="This imaging series is corrected for the haemodynamic response.",
    )

    ophys_module = nwb.create_processing_module(name="ophys", description="optical physiology processed data")

    ophys_module.add(deltaF_series)

    io.write(nwb)

    return nwbfile


@pytest.fixture
def output_dir(tmp_path_factory):
    """Create a temporary directory for output."""
    return str(tmp_path_factory.mktemp("output"))


@pytest.fixture
def meta_yaml(tmp_path_factory, partial=False):
    tmpfile = tmp_path_factory.mktemp("data") / "test_meta.yml"

    metadata = {
        "subject_id": "testyaml",
        "sex": "Male",
        "genotype": "Wt",
        "species": "Mus musculus",
        "strain": "C57/B6",
        "dob": "2025-01-11",
        "session_description": "This is but a test.",
        "experimenter": "John Doe",
        "lab": "Doe lab",
        "institution": "University of Someplace with lots of funding",
    }

    if partial:
        metadata.pop("lab")
        metadata.pop("genotype")

    with open(tmpfile, "w") as fp:
        yaml.safe_dump(metadata, fp)

    return str(tmpfile)


@pytest.fixture
def meta_json(tmp_path_factory, partial=False):
    tmpfile = tmp_path_factory.mktemp("data") / "test_meta.json"

    metadata = {
        "subject_id": "testjson",
        "sex": "Male",
        "genotype": "Wt",
        "species": "Mus musculus",
        "strain": "C57/B6",
        "dob": "2025-01-11",
        "session_description": "This is but a test.",
        "experimenter": "John Doe",
        "lab": "Doe lab",
        "institution": "University of Someplace with lots of funding",
    }

    if partial:
        metadata.pop("lab")
        metadata.pop("genotype")

    with open(tmpfile, "w") as fp:
        json.dump(metadata, fp)

    return str(tmpfile)



#  Copyright (c) 2024 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import csv
import typing
from collections import OrderedDict

import h5py
import numpy.typing as npt
import xmltodict
import zarr
from dask import array as da
from pynwb import NWBHDF5IO
from pynwb import NWBFile


@typing.overload
def read_nwb(path: str, mode: str = "a") -> NWBFile: ...  # pyright: ignore[reportOverlappingOverload]


@typing.overload
def read_nwb(path: str, mode: str = "a", return_io: bool = True) -> tuple[NWBFile, NWBHDF5IO]: ...


def read_nwb(path: str, mode: str = "a", return_io: bool = False) -> NWBFile | tuple[NWBFile, NWBHDF5IO]:
    """Read an NWB file.

    Args:
        path (str): Path to the NWB file.
        mode (str, optional): File read mode (i.e. read/write/append). Defaults to "a".
        return_io (bool, optional): Return IO object alongside the NWB file object. Defaults to False.

    Returns:
        NWBFile: NWB file object.
        NWBHDF5IO: IO object (if return_io=True).
    """
    io = NWBHDF5IO(path, mode=mode)
    nwbfile = io.read()
    if return_io:
        return nwbfile, io
    return nwbfile


def write_nwb(path: str, nwbfile: NWBFile, mode: str = "w", io: NWBHDF5IO = None, **kwargs) -> None:
    """Write an NWB file.

    Args:
        path (str): Path to the NWB file.
        nwbfile (NWBFile): NWB file object.
        mode (str, optional): File write mode (i.e. write/append). Defaults to "w".
        **kwargs: Parameters passed to NWBHDF5IO.write.
    """
    if io:
        return io.write(nwbfile, **kwargs)
    with NWBHDF5IO(path, mode=mode) as io:
        return io.write(nwbfile, **kwargs)


def read_h5(path: str) -> h5py.File:
    """Read an HDF5 file.

    Args:
        path (str): Path to the HDF5 file.

    Returns:
        h5py.File: HDF5 file object.
    """
    return h5py.File(path, "r")


def write_h5(path: str, data: dict, compression: str = "lzf", attributes: dict = {}) -> str:
    """Write a dictionary to an HDF5 file.

    Args:
        path (str): Path to the HDF5 file.
        data (dict): Dictionary containing datasets to write in {'dataset_name': data_array} format.
        compression (str, optional): Compression method for the datasets. Defaults to "lzf".
        attributes (dict, optional): Attributes to write to the HDF5 file. Defaults to {}.

    Returns:
        str: Path to the written HDF5 file.

    Example:
        >>> data = {"dataset1": np.array([1, 2, 3]), "dataset2": np.array([[1, 2], [3, 4]])}
        >>> write_h5("output.h5", data)
    """
    with h5py.File(path, "w") as h5file:
        for key, value in data.items():
            h5file.create_dataset(key, data=value, compression=compression)
        if attributes:
            h5file.attrs.update(attributes)
    return path


def store_interim(
    array: da.Array | npt.ArrayLike,
    interim_path: str,
    compute: bool = True,
    chunks: int = 500,
) -> zarr.core.Array:
    """Store an array in an interim Zarr file.

    Args:
        array (Dask or Numpy Array): Dask or Numpy array to persist on disk.
        interim_path (str): Path to the interim Zarr file. The .zarr extension is added automatically.
        compute (bool, optional): Whether to compute the array before storing, applies only to Dask arrays. Defaults to True.
        chunks (int, optional): Chunk size for the Zarr file. Defaults to 500.

    Returns:
        Zarr Array: Persistent Zarr array object
    """
    if not interim_path.endswith(".zarr"):
        interim_path += ".zarr"

    if isinstance(array, da.Array):
        z_interim = zarr.open_array(
            interim_path,
            shape=array.shape,
            dtype=array.dtype,
            chunks=(chunks, array.shape[1], array.shape[2]),
        )
        return array.store(z_interim, return_stored=True, compute=compute)  # type: ignore

    zarr.save(interim_path, array)
    return zarr.load(interim_path)


def read_points(path: str) -> dict[str, tuple[float, float]]:
    """Read a landmark points file.

    Args:
        path (str): Path to the points file.

    Returns:
        dict[str, tuple[float, float]]: Dictionary with the landmark names as keys and their x-y coordinates
    """
    if path.endswith(".xml") or path.endswith(".points"):
        return _read_fiji_points(path)
    if path.endswith(".csv"):
        return _read_csv_points(path)
    raise ValueError("Unsupported file format.")


def _read_fiji_points(path: str) -> dict[str, tuple[float, float]]:
    """Read a FIJI landmark points file.

    Args:
        path (str): Path to the points file.

    Returns:
        dict[str, tuple[float, float]]: Dictionary with the landmark names as keys and their x-y coordinates
    """
    with open(path) as fp:
        points = xmltodict.parse(fp.read())
        points = OrderedDict(
            {
                point["@name"]: (float(point["@y"]), float(point["@x"]))
                for point in points["namedpointset"]["pointworld"]
            }
        )

    return points


def _read_csv_points(path: str) -> dict[str, tuple[float, float]]:
    """Read a CSV file with landmark points.

    Args:
        path (str): Path to the CSV file.

    Returns:
        dict[str, tuple[float, float]]: Dictionary with the landmark names as keys and their x-y coordinates
    """
    with open(path) as fp:
        csv_reader = csv.DictReader(fp)
        points = OrderedDict({row["landmark"]: (float(row["y"]), float(row["x"])) for row in csv_reader})

    return points


def write_points(path: str, points: dict[str, tuple[float, float]]) -> None:
    """Write a dictionary of landmark points to a CSV file.

    Args:
        path (str): Path to output CSV file.
        points (dict[str, tuple[float, float]]): Dictionary with the landmark names as keys and their x-y coordinates
    """
    if not path.endswith(".csv"):
        path += ".csv"
    with open(path, "w") as fp:
        csv_writer = csv.DictWriter(fp, fieldnames=["landmark", "y", "x"])
        csv_writer.writeheader()
        for landmark, (y, x) in points.items():
            csv_writer.writerow({"landmark": landmark, "y": y, "x": x})



#  Copyright (c) 2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
"""Module for quality assurance (QA) functions in the mesoscopy preprocessing pipeline."""

import numpy.typing as npt
import plotly.express as px
import plotly.graph_objects as go


def plot_timestamps(timestamps: list | npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a sequence of timestamps as a line plot using Plotly.

    Args:
        timestamps (list | npt.NDArray): Sequence of timestamp values to plot.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Frame index"}},
            yaxis={"title": {"text": "Timestamp"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=timestamps, mode="lines", name="Timestamp"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_raw_timeseries(raw_timeseries: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a raw timeseries signal using Plotly.

    Args:
        raw_timeseries (npt.NDArray): The raw timeseries data to plot, as a NumPy array.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Frame index"}},
            yaxis={"title": {"text": "Signal"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=raw_timeseries, mode="markers", name="Signal"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_raw_histogram(data: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a histogram of raw data using Plotly.

    Args:
        data (npt.NDArray): The raw data to plot as a histogram, as a NumPy array.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Signal"}},
            yaxis={"title": {"text": "Count"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Histogram(x=data))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_channels_timeseries(
    gcamp_channel: npt.NDArray, isosb_channel: npt.NDArray, as_html: bool = False
) -> str | go.Figure:
    """Plots the timeseries of GCaMP and Isosb channels.

    Args:
        gcamp_channel (npt.NDArray): The GCaMP channel timeseries data
        isosb_channel (npt.NDArray): The Isosb channel timeseries data
        as_html (bool, optional): If True, returns the plot as an HTML string.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Frame index"}},
            yaxis={"title": {"text": "Signal"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=gcamp_channel, mode="lines", name="GCaMP channel"))
    fig.add_trace(go.Scatter(y=isosb_channel, mode="lines", name="Isosb channel"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_projection(data: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a channel projection using Plotly.

    Args:
        data (npt.NDArray): The projection data to plot as a NumPy array.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    if as_html:
        return px.imshow(data).to_html(full_html=False)
    return px.imshow(data)


def plot_frame_ids(gcamp_ids: npt.NDArray, isosb_ids: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots the frame IDs of GCaMP and Isosb channels.

    Args:
        gcamp_ids (npt.NDArray): The GCaMP channel frame IDs.
        isosb_ids (npt.NDArray): The Isosb channel frame IDs.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Processed frame index"}},
            yaxis={"title": {"text": "Original frame index"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=gcamp_ids, mode="lines", name="GCaMP channel IDs"))
    fig.add_trace(go.Scatter(y=isosb_ids, mode="lines", name="Isosb channel IDs"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_filter_pie(gcamp_ids: npt.NDArray, isosb_ids: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a pie chart of the frame IDs for GCaMP and Isosb channels.

    Args:
        gcamp_ids (npt.NDArray): The GCaMP channel frame IDs.
        isosb_ids (npt.NDArray): The Isosb channel frame IDs.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    labels = ["GCaMP frame IDs", "Isosb frame IDs"]
    values = [len(gcamp_ids), len(isosb_ids)]
    fig = go.Figure(
        go.Pie(labels=labels, values=values),
        layout=go.Layout(
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        ),
    )
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_channels_dff(gcamp_channel: npt.NDArray, isosb_channel: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots the ∆F/F time series of GCaMP and Isosb channels.

    Args:
        gcamp_channel (npt.NDArray): The GCaMP channel ∆F/F time series data.
        isosb_channel (npt.NDArray): The Isosb channel ∆F/F time series data.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Frame index"}},
            yaxis={"title": {"text": "∆F/F"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=gcamp_channel, mode="lines", name="GCaMP channel ∆F/F"))
    fig.add_trace(go.Scatter(y=isosb_channel, mode="lines", name="Isosb channel ∆F/F"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_corrected_dff(data: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots the corrected ∆F/F time series.

    Args:
        data (npt.NDArray): The corrected ∆F/F time series data.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"title": {"text": "Frame index"}},
            yaxis={"title": {"text": "∆F/F"}},
            margin={"l": 20, "r": 20, "t": 20, "b": 20},
        )
    )

    fig.add_trace(go.Scatter(y=data, mode="lines", name="Corrected ∆F/F"))
    if as_html:
        return fig.to_html(full_html=False)
    return fig


def plot_frame(data: npt.NDArray, as_html: bool = False) -> str | go.Figure:
    """Plots a single frame of data using Plotly.

    Args:
        data (npt.NDArray): The frame data to plot, as a NumPy array.
        as_html (bool, optional): If True, returns the plot as an HTML string. If False, returns a Plotly Figure object.

    Returns:
        str | go.Figure: The plot as an HTML string if `as_html` is True, otherwise a Plotly Figure object.
    """
    if as_html:
        return px.imshow(data).to_html(full_html=False)
    return px.imshow(data)



#  Copyright (c) 2022 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import os
import shutil
import time
import typing

import click
import dask
import numpy as np
from dask import array as da
from pynwb.image import ImageSeries

import mesoscopy.preprocess.compute as calc
from mesoscopy import io
from mesoscopy import timer


@click.command(name="preprocess")
@click.argument(
    "path",
    type=click.Path(exists=True),
)
@click.option(
    "-o",
    "--out_dir",
    type=click.Path(dir_okay=True),
    default="./",
    help="Output directory for preprocessed recording.",
)
@click.option("--chunks", default=100, help="Number of chunks to load in memory.")
@click.option(
    "--crop",
    default=0,
    help="Number of pixels to crop from the edges of the recording.",
)
@click.option("--bins", default=2, help="Recording pixel binning factor.")
@click.option(
    "--channel-means-only",
    is_flag=True,
    show_default=True,
    default=False,
    help="Extract the channel means and exit without extracting a delta F series.",
)
@click.option(
    "--use-means",
    is_flag=True,
    show_default=True,
    default=False,
    help="Separate channels using means histogram instead of standard deviation.",
)
@click.option(
    "--flip-channels",
    is_flag=True,
    show_default=True,
    default=False,
    help="Flip channel order.",
)
@click.option("--interim_dir", type=click.Path(dir_okay=True), default="interim/")
@click.option(
    "--skip-start",
    default=None,
    type=int,
    help="Number of frames to skip at the start of the recording.",
)
@click.option(
    "--skip-end",
    default=None,
    type=int,
    help="Number of frames to skip at the end of the recording.",
)
def preprocess_cmd(
    **kwargs: typing.Any,
) -> None:
    """This command will preprocess a single session dual-channel mixed recording to extract a deltaF signal corrected for the haemodynamic response."""
    run_preprocessing(**kwargs)


def run_preprocessing(
    path: str,
    out_dir: str,
    chunks: int = 100,
    crop: int = 0,
    bins: int = 2,
    channel_means_only: bool = False,
    use_means: bool = False,
    flip_channels: bool = False,
    interim_dir: str = "interim/",
    skip_start: int | None = None,
    skip_end: int | None = None,
) -> None:
    """Preprocessing to extract deltaF from a single session dual-channel mixed recording.

    Preprocessing separates the two channels, applies the haemodynamic correction,
    and extracts the delta F signal.

    Args:
        raw_path (str): Path to the raw recording HDF5 or NWB file.
        out_dir (str): Path to the output directory.
        chunks (int, optional): Number of chunks to load in memory. Defaults to 100.
        crop (int, optional): Number of pixels to crop from the edges of the recording. Defaults to 0.
        bins (int, optional): Recording pixel binning factor. Defaults to 2.
        channel_means_only (bool, optional): Extract the channel means and exit without extracting a delta F series. Defaults to False.
        use_means (bool, optional): Use means histogram instead of standard deviation to separate channels. Defaults to False.
        flip_channels (bool, optional): Flip extracted channel order. Defaults to False.
        interim_dir (str, optional): Path to the interim directory. Defaults to "interim/".
        skip_start (int, optional): Number of frames to skip at the start of the recording. Defaults to None.
        skip_end (int, optional): Number of frames to skip at the end of the recording. Defaults to None.
    """
    click.echo(f"Preprocessing file {path}.")

    preprocessing_start = time.time()

    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(interim_dir, exist_ok=True)

    click.echo("Loading data...")

    # Determine whether we're working with an NWB file
    nwb = True if path.endswith(".nwb") else False

    session_id, d, ts = load_raw(path, nwb=nwb)

    if skip_start and skip_start != 0:
        skip_start = skip_start - 1

    # If odd or even skips, add one to make them even to avoid strangeness
    if skip_start and skip_start % 2 != 0:
        skip_start += 1

    if skip_end and skip_end % 2 != 0:
        skip_end += 1

    if skip_end:
        skip_end = -abs(skip_end)

    if skip_start or skip_end:
        d = d[skip_start:skip_end]
        ts = ts[skip_start:skip_end]

    raw_frames = da.from_array(d, chunks="auto")
    if chunks > 0:
        raw_frames = raw_frames.rechunk(chunks=(chunks, d.shape[1], d.shape[2]))

    if crop > 0:
        raw_frames = raw_frames[:, crop:-crop, crop:-crop]
        click.echo(f"Cropping to shape {raw_frames.shape}")

    with timer.Timer(
        f"Binning frames to shape {raw_frames.shape[1] // bins} by {raw_frames.shape[2] // bins} ({bins}x{bins})"
    ):
        binned_frames = calc.bin_array(
            raw_frames,
            bins=bins,
            interim_dir=interim_dir,
            session_id=session_id,
        )

    del raw_frames

    # Channel separation
    # Get the global mean and std values for each frame
    with timer.Timer("Calculating frame means & standard deviations"):
        frame_means, frame_stds = calc.frame_statistics(binned_frames)

    # Generate the channel separation filters
    with timer.Timer("Generating channel separation filters"):
        gcamp_filter, isosb_filter = calc.channel_separation_filters(
            frame_means,
            frame_stds,
            use_means=use_means,
            flip_channels=flip_channels,
        )

    with timer.Timer("Calculating channel mean timeseries"):
        gcamp_mean, isosb_mean = dask.compute(
            binned_frames[gcamp_filter].mean(axis=(1, 2), dtype=np.float32),
            binned_frames[isosb_filter].mean(axis=(1, 2), dtype=np.float32),
        )

    if channel_means_only:
        np.savetxt(
            os.path.join(out_dir, f"{session_id}_gcamp_mean.txt"),
            gcamp_mean,
            fmt="%.4f",
        )
        np.savetxt(
            os.path.join(out_dir, f"{session_id}_isosb_mean.txt"),
            isosb_mean,
            fmt="%.4f",
        )
        click.echo("Channel means saved as txt files. Exiting.")
        return

    gcamp_channel = binned_frames[gcamp_filter]
    isosb_channel = binned_frames[isosb_filter]

    with timer.Timer("Generating projection images per channel"):
        gcamp_projections = calc.projections(gcamp_channel)
        isosb_projections = calc.projections(isosb_channel)

    # Calculate the dff per channel using a rolling baseline (mean in a 30s window)
    window_width = 30 * 25

    if window_width > binned_frames.shape[0]:
        click.echo(
            f"WARNING: Default window width is larger than the number of frames. Setting window width to {binned_frames.shape[0] // 4}."
        )
        window_width = binned_frames.shape[0] // 4

    with timer.Timer("Calculating dF/F per channel"):
        gcamp_dff = calc.rolling_dff(
            gcamp_channel,
            window_width=window_width,
            channel_name="gcamp",
            interim_dir=interim_dir,
            session_id=session_id,
        )
        isosb_dff = calc.rolling_dff(
            isosb_channel,
            window_width=window_width,
            channel_name="isosb",
            interim_dir=interim_dir,
            session_id=session_id,
        )

    with timer.Timer("Calculating mean ∂F per frame per channel"):
        gcamp_signal_mean, isosb_signal_mean = da.compute(gcamp_dff.mean(axis=(1, 2)), isosb_dff.mean(axis=(1, 2)))

    # Max common index (to avoid array overflow)
    if len(gcamp_mean) != len(isosb_mean):
        click.echo("WARNING: GCaMP & Isosb channels have mismatching indexes")
    max_idx = min(len(gcamp_mean), len(isosb_mean))

    click.echo("Extracting corrected F signal (gcamp - isosb)...")

    with timer.Timer("Calculating F signal"):
        f_signal = da.subtract(
            gcamp_dff[:max_idx],
            isosb_dff[:max_idx],
        )

    with timer.Timer("Calculating mean F timeseries"):
        f_mean_timeseries = f_signal.mean(axis=(1, 2))

    with timer.Timer("Generating maximum intensity projection for F signal"):
        f_maxip = calc.projections(f_signal).get("maxip")

    timestamps = da.from_array(np.array(ts[gcamp_filter[:max_idx]], dtype="S25"), chunks="auto")

    outpath = out_dir + os.sep + session_id + "_preprocessed.h5"
    with timer.Timer(f"Writing data to {outpath}"):
        # da.to_hdf5(
        io.write_h5(
            path=outpath,
            data={
                "/F": f_signal,
                "/timestamps": timestamps,
                "/qa/frame_means_timeseries": frame_means,
                "/qa/frame_stds_timeseries": frame_stds,
                "/qa/gcamp_mean_timeseries": gcamp_mean[:max_idx],
                "/qa/isosb_mean_timeseries": isosb_mean[:max_idx],
                "/qa/gcamp_dff_timeseries": gcamp_signal_mean[:max_idx],
                "/qa/isosb_dff_timeseries": isosb_signal_mean[:max_idx],
                "/qa/gcamp_filter": gcamp_filter[:max_idx],
                "/qa/isosb_filter": isosb_filter[:max_idx],
                "/qa/gcamp_mean_projection": gcamp_projections.get("mean"),
                "/qa/gcamp_std_projection": gcamp_projections.get("std"),
                "/qa/gcamp_maxip_projection": gcamp_projections.get("maxip"),
                "/qa/isosb_mean_projection": isosb_projections.get("mean"),
                "/qa/isosb_std_projection": isosb_projections.get("std"),
                "/qa/isosb_maxip_projection": isosb_projections.get("maxip"),
                "/qa/f_mean_timeseries": f_mean_timeseries,
                "/qa/f_maxip": f_maxip,
            },
            compression="lzf",
        )

    preprocessing_end = time.time()
    click.echo(f"Preprocessing took a total of {(preprocessing_end - preprocessing_start) / 60} mins.")

    if nwb:
        click.echo("Updating NWB file...")
        update_nwb(path, outpath)
        click.echo(f"Updated NWB file at {path}")

    click.echo("Cleaning up...")
    shutil.rmtree(interim_dir)


def load_raw(raw_path: str, nwb: bool = False) -> tuple[str, da.Array | np.ndarray, da.Array | np.ndarray]:
    """Load raw imaging data from an HDF5 or NWB file.

    Args:
        raw_path (str): Path to the raw recording HDF5 or NWB file.
        nwb (bool, optional): Whether the file is an NWB file. Defaults to False.

    Returns:
        tuple[str, da.Array | np.ndarray, da.Array | np.ndarray]: Session ID, imaging data, and timestamps.
    """
    if nwb:
        nwbfile = io.read_nwb(raw_path)

        session_id = nwbfile.identifier

        imaging_data = nwbfile.acquisition["DualChannelImagingSeries"].data[:]
        timestamps = nwbfile.acquisition["DualChannelImagingSeries"].timestamps[:]
    else:
        session_id = raw_path.split("/")[-1].replace(".h5", "")

        # Lazy-load the data into a dask array
        f_raw = io.read_h5(raw_path)
        imaging_data = f_raw["/frames"]
        timestamps = f_raw["/timestamps"]

    return session_id, imaging_data, timestamps


def update_nwb(nwb_path: str, h5_path: str) -> None:
    """Update an NWB file with a delta F imaging series stored in an HDF5 file.

    Creates a link between the NWB file and the HDF5 file. See https://pynwb.readthedocs.io/en/stable/tutorials/advanced_io/linking_data.html.

    Args:
        nwb_path (str): Path to NWB file.
        h5_path (str): Path to HDF5 file containing the delta F imaging series.
    """
    f = io.read_h5(h5_path)
    nwbfile, nwbio = io.read_nwb(nwb_path, return_io=True)
    deltaF_series = ImageSeries(
        name="DeltaFSeries",
        data=f["/F"],
        timestamps=f["/timestamps"],
        unit="df/f",
        description="dF/F widefield cortical imaging series.",
        comments="This imaging series is corrected for the haemodynamic response.",
    )

    ophys_module = nwbfile.create_processing_module(name="ophys", description="optical physiology processed data")

    ophys_module.add(deltaF_series)

    io.write_nwb(nwb_path, nwbfile, io=nwbio)

    return nwbfile



import pathlib

import numpy as np
import pytest
import zarr
from click.testing import CliRunner
from dask import array as da

import mesoscopy
from mesoscopy import preprocess
from mesoscopy.preprocess import compute


def test_load_raw_h5(raw_h5):
    session_id, imaging_data, timestamps = preprocess.load_raw(raw_h5, nwb=False)
    assert session_id == "preproc_test"
    assert imaging_data.shape == (600, 40, 40)
    assert len(timestamps) == 600


def test_load_raw_nwb(nwbfile):
    session_id, imaging_data, timestamps = preprocess.load_raw(nwbfile, nwb=True)
    assert session_id == "session_1234"
    assert imaging_data.shape == (600, 40, 40)
    assert len(timestamps) == 600


def test_channel_qa():
    pass


def test_nwb_link(nwbfile, preproc_h5):
    nwb = preprocess.update_nwb(nwbfile, preproc_h5)
    assert nwb.processing["ophys"]["DeltaFSeries"].data
    assert nwb.processing["ophys"]["DeltaFSeries"].data.shape == (300, 40, 40)
    assert nwb.processing["ophys"]["DeltaFSeries"].timestamps


def test_binning(output_dir):
    array = compute.bin_array(
        array=da.random.random((300, 40, 40)),
        bins=2,
        interim_dir=output_dir,
        session_id="null",
    )
    assert array.shape == (300, 20, 20)


def test_separate_channels(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 40, 40))
    mock_isosb = np.random.normal(65, 50, size=(300, 40, 40))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, isosb_filter = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=False,
        flip_channels=False,
    )

    assert mock_dual_channel[gcamp_filter].shape[0] == 300
    assert mock_dual_channel[isosb_filter].shape[0] == 300
    assert (isosb_filter == random_idx).all()


def test_separate_channels_use_means(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(200, 50, size=(300, 40, 40))
    mock_isosb = np.random.normal(65, 50, size=(300, 40, 40))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, isosb_filter = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=True,
        flip_channels=False,
    )

    assert mock_dual_channel[gcamp_filter].shape[0] == 300
    assert mock_dual_channel[isosb_filter].shape[0] == 300
    assert (isosb_filter == random_idx).all()


def test_separate_channels_flip_channels(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 40, 40))
    mock_isosb = np.random.normal(65, 50, size=(300, 40, 40))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, isosb_filter = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=False,
        flip_channels=True,
    )

    assert mock_dual_channel[gcamp_filter].shape[0] == 300
    assert mock_dual_channel[isosb_filter].shape[0] == 300
    assert (gcamp_filter == random_idx).all()


def test_channel_dff(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 20, 20))
    mock_isosb = np.random.normal(65, 50, size=(300, 20, 20))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, _ = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=False,
        flip_channels=False,
    )

    dff = compute.rolling_dff(
        array=mock_dual_channel[gcamp_filter],
        interim_dir=output_dir,
        session_id="null",
        window_width=10,
    )
    assert dff.shape == (300, 20, 20)


def test_channel_dff_invalid_window(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 20, 20))
    mock_isosb = np.random.normal(65, 50, size=(300, 20, 20))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, isosb_filter = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=False,
        flip_channels=False,
    )

    with pytest.raises(ValueError):
        compute.rolling_dff(
            array=mock_dual_channel[gcamp_filter],
            interim_dir=output_dir,
            session_id="null",
            window_width=700,
        )


def test_channel_dff_insert(output_dir, random_idx):
    # Generate mock dual-channel imaging data
    mock_gcamp = np.random.normal(70, 200, size=(300, 20, 20))
    mock_isosb = np.random.normal(65, 50, size=(300, 20, 20))

    # Merge the two channels in random order
    mock_dual_channel = np.insert(mock_gcamp, random_idx - np.arange(len(random_idx)), mock_isosb, axis=0)

    gcamp_filter, _ = compute.channel_separation_filters(
        frame_means=mock_dual_channel.mean(axis=(1, 2)),
        frame_stds=mock_dual_channel.std(axis=(1, 2)),
        use_means=False,
        flip_channels=False,
    )

    _ = compute.rolling_dff(
        array=mock_dual_channel[gcamp_filter],
        interim_dir=output_dir,
        session_id="null",
        window_width=10,
    )

    # Assert insertion happens in a block
    f0 = zarr.load(pathlib.Path(output_dir) / "null_null_f0_appended.zarr")
    assert da.all(da.equal(f0[0], f0[1]))
    assert da.all(da.equal(f0[-2], f0[-1]))


def test_preprocess_h5(raw_h5, output_dir):
    preprocess.run_preprocessing(path=raw_h5, out_dir=output_dir, interim_dir=output_dir)


def test_preprocess_h5_crop(raw_h5, output_dir):
    preprocess.run_preprocessing(path=raw_h5, out_dir=output_dir, interim_dir=output_dir, crop=5)


def test_preprocess_h5_frameskip(raw_h5, output_dir):
    # Skip odd frames
    preprocess.run_preprocessing(
        path=raw_h5,
        out_dir=output_dir,
        interim_dir=output_dir,
        skip_start=5,
        skip_end=5,
    )

    # Skip even frames
    preprocess.run_preprocessing(
        path=raw_h5,
        out_dir=output_dir,
        interim_dir=output_dir,
        skip_start=50,
        skip_end=50,
    )


def test_preprocess_h5_bin(raw_h5, output_dir):
    preprocess.run_preprocessing(path=raw_h5, out_dir=output_dir, interim_dir=output_dir, bins=4)


def test_preprocess_h5_channelmeansonly(raw_h5, output_dir):
    preprocess.run_preprocessing(path=raw_h5, out_dir=output_dir, interim_dir=output_dir, channel_means_only=True)


def test_preprocess_nwb(nwbfile, output_dir):
    preprocess.run_preprocessing(path=nwbfile, out_dir=output_dir, interim_dir=output_dir)


def test_preprocess_nwb_crop(nwbfile, output_dir):
    preprocess.run_preprocessing(path=nwbfile, out_dir=output_dir, interim_dir=output_dir, crop=5)


def test_preprocess_nwb_frameskip_odd(nwbfile, output_dir):
    # Skip odd frames
    preprocess.run_preprocessing(
        path=nwbfile,
        out_dir=output_dir,
        interim_dir=output_dir,
        skip_start=5,
        skip_end=5,
    )


def test_preprocess_nwb_frameskip_even(nwbfile, output_dir):
    # Skip even frames
    preprocess.run_preprocessing(
        path=nwbfile,
        out_dir=output_dir,
        interim_dir=output_dir,
        skip_start=50,
        skip_end=50,
    )


def test_preprocess_nwb_bin(nwbfile, output_dir):
    preprocess.run_preprocessing(path=nwbfile, out_dir=output_dir, interim_dir=output_dir, bins=4)


def test_preprocess_nwb_channelmeansonly(nwbfile, output_dir):
    preprocess.run_preprocessing(
        path=nwbfile,
        out_dir=output_dir,
        interim_dir=output_dir,
        channel_means_only=True,
    )


def test_cmd_parity(raw_h5, output_dir):
    """Check if click preprocess cmd caller works"""
    runner = CliRunner()
    result = runner.invoke(
        mesoscopy.cli,
        args=f"preprocess --out_dir {output_dir} --interim_dir {output_dir} {raw_h5}",
    )
    assert result.exit_code == 0



#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""Video conversion utilities."""

import datetime
import os
import typing

import h5py as h5
import numpy as np
import pandas as pd
import pims
from tqdm import tqdm

import mesoscopy.convert.hdf5 as conv_h5


def to_hdf5(
    input_path: str,
    out_dir: str,
    ts_path: str = "",
    ts_delimiter: str = ",",
    ts_column: str | int = 0,
    ts_has_header: bool = False,
    **kwargs: typing.Any,
) -> str:
    video_data = pims.as_gray(pims.PyAVVideoReader(input_path))

    if ts_path:
        header_row = 0 if ts_has_header else None
        ts_df = pd.read_csv(
            ts_path,
            delimiter=ts_delimiter,
            header=header_row,
        )

        timestamps = ts_df.loc[:, ts_column].values
    else:
        timestamps = [str(datetime.datetime.now()).replace(" ", "T") for _ in range(len(video_data))]

    fname = input_path.split(os.sep)[-1].split(".")[0]
    out_path = f"{out_dir}{os.sep}{fname}.h5"

    with h5.File(out_path, "w") as f:
        f.create_dataset("timestamps", data=timestamps)
        f.create_dataset(
            "frames",
            shape=(len(video_data), *video_data[0].shape),
            chunks=(100, video_data[0].shape[0], video_data[0].shape[1]),
            dtype=np.int8,
        )

        for idx, frame in tqdm(
            enumerate(video_data),
            desc="Converting video to HDF5",
            total=len(video_data),
        ):
            f["frames"][idx] = frame

    return out_path


def to_nwb(input_path: str, **kwargs: typing.Any) -> str:
    h5_path = to_hdf5(input_path, **kwargs)

    return conv_h5.to_nwb(h5_path, **kwargs)



#  Copyright (c) 2022 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""Main entry point to the mesoscopy CLI"""

import os

import click
import h5py
from matplotlib import pyplot as plt

import mesoscopy.register as reg
from mesoscopy import convert
from mesoscopy import export
from mesoscopy import inspect
from mesoscopy import postprocess
from mesoscopy import preprocess
from mesoscopy import process
from mesoscopy import report
from mesoscopy.__about__ import __version__


@click.group()
@click.version_option(__version__)
def cli():
    """Widefield calcium imaging analysis pipeline."""


@cli.command()
@click.argument("path", type=click.Path(exists=True))
@click.argument("out_dir", type=click.Path(dir_okay=True))
@click.option("--index", default=0, show_default=True, help="Index of frame to sample.")
@click.option("--crop", default=0)
@click.option("--vmin", type=float, default=0)
@click.option("--vmax", type=float, default=255)
@click.option("--key", type=str, default=None, help="Activity column")
def sample(path, out_dir, index, crop=0, vmin=0, vmax=255, key="frames"):
    """Sample an image frame from an HDF5 file and export it as a PNG."""
    click.echo(f"Sampling {path} at index {index}.")

    f = h5py.File(path)
    d = f[f"/{key}"]

    os.makedirs(out_dir, exist_ok=True)
    outpath = out_dir + os.sep + path.split("/")[-1].replace(".h5", f"_sample-{index}.png")

    if crop > 0:
        plt.imsave(outpath, d[index, crop:-crop, crop:-crop], vmin=vmin, vmax=vmax)
    else:
        plt.imsave(outpath, d[index], vmin=vmin, vmax=vmax, cmap="jet")
    click.echo(f"Saved sample at {outpath}")


cli.add_command(preprocess.preprocess_cmd)
cli.add_command(process.process_cmd)
cli.add_command(postprocess.postprocess_cmd)
cli.add_command(reg.register_cmd)
cli.add_command(convert.convert_cmd)
cli.add_command(inspect.inspect_cmd)
cli.add_command(export.export_cmd)
cli.add_command(report.report_cmd)



#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
"""Export functions for mesoscopy-generated files."""

import typing

import click

import mesoscopy.export.nwb as exp_nwb


@click.group("export")
def export_cmd() -> None:
    """Export mesoscopy-generated files."""


@export_cmd.command("nwb")
@click.argument("nwb_path", type=click.Path(exists=True, dir_okay=False, path_type=str))
@click.option(
    "--out-path",
    type=click.Path(dir_okay=False, path_type=str),
    default="",
    help="Path to save the exported NWB file. If not provided, the output file will be named as the input file with '_export.nwb' appended.",
)
def export_nwb_cmd(**kwargs: typing.Any) -> None:
    """Create a sharable copy of an NWB file by resolving external data links."""
    click.echo("Exporting NWB file...")
    export_path = export_nwb(**kwargs)
    click.echo(f"NWB file exported to {export_path}")


def export_nwb(nwb_path: str, out_path: str = "") -> str:
    """Create a sharable copy of an NWB file by resolving external data links.

    Args:
        nwb_path (str): Path to the source NWB file.
        out_path (str, optional): Path to save the exported NWB file. If not provided, the output file will be named as the input file with '_export.nwb' appended. Defaults to "".

    Returns:
        str: Path to the exported NWB file.
    """
    return exp_nwb.export_standalone(nwb_path, out_path)



import click
from matplotlib import pyplot as plt


def plot_line(array, outpath, message=None):
    plt.clf()
    plt.plot(array)
    plt.savefig(outpath)
    if message:
        click.echo(message)


def plot_lines(arrays, outpath, message=None):
    plt.clf()
    for array in arrays:
        plt.plot(array)
    plt.savefig(outpath)
    if message:
        click.echo(message)


def plot_hist(array, outpath, message=None):
    plt.clf()
    plt.hist(array)
    plt.savefig(outpath)
    if message:
        click.echo(message)


def plot_frame(array, outpath, message=None):
    plt.clf()
    plt.imsave(outpath, array)
    if message:
        click.echo(message)


def plot_scatter(x, y, outpath, labels=[], message=None):
    plt.clf()
    plt.scatter(x, y)
    if labels:
        plt.legend(labels)
    plt.savefig(outpath)
    if message:
        click.echo(message)


def plot_scatters(xs, ys, outpath, labels=[], message=None):
    plt.clf()
    for x, y in zip(xs, ys):
        plt.scatter(x, y)
    if labels:
        plt.legend(labels)
    plt.savefig(outpath)
    if message:
        click.echo(message)



#  Copyright (c) 2024 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import time

import click
import numpy as np
from skimage import transform as trf


def landmarks_affine(
    deltaf_series: np.ndarray,
    recording_landmarks: dict,
    template_landmarks: dict,
    crop_x: int = 0,
    crop_y: int = 0,
) -> tuple[np.ndarray, trf.ProjectiveTransform]:
    """Warp a DeltaF/F series to match a template using anatomical landmarks.

    Args:
        deltaf_series (np.ndarray): DeltaF/F series.
        recording_landmarks (dict): Recording landmarks.
        template_landmarks (dict): Template landmarks.
        crop_x (int, optional): Crop x-axis. Defaults to 0.
        crop_y (int, optional): Crop y-axis. Defaults to 0.
        qa_dir (str, optional): Directory to save QA plots. Defaults to "".
        session_id (str, optional): Session ID. Defaults to "".

    Returns:
        tuple[np.ndarray, np.ndarray]: Registered DeltaF/F series and affine transformation matrix.
    """
    template = np.array(list(template_landmarks.values()), dtype=np.float32)
    recording = np.array(list(recording_landmarks.values()), dtype=np.float32)

    click.echo("Estimating transform...")
    start = time.time()
    tform = trf.estimate_transform("affine", template, recording)
    end = time.time()
    click.echo(f"Transform estimated in {end - start} s")

    warped_ = []
    with click.progressbar(range(deltaf_series.shape[0]), label="Registering recording to template...") as frame_ids:
        for idx in frame_ids:
            if crop_x > 0 or crop_y > 0:
                warped_.append(trf.warp(deltaf_series[idx, :crop_y, :crop_x], tform.inverse, order=3))
            else:
                warped_.append(trf.warp(deltaf_series[idx, :, :], tform.inverse, order=3))
    warped = np.array(warped_)

    return warped, tform



#  Copyright (c) 2024 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from collections import OrderedDict

import magicgui.widgets as mgw
import napari
import numpy as np
import numpy.typing as npt
from dask import array as da

COLOR_CYCLE = [
    "#a6cee3",
    "#1f78b4",
    "#b2df8a",
    "#33a02c",
    "#fb9a99",
    "#e31a1c",
    "#fdbf6f",
    "#ff7f00",
    "#cab2d6",
]


def mark_landmarks(
    maxip_image: npt.NDArray | da.Array, alt_image: npt.NDArray | da.Array | None, template_landmarks: dict = {}
) -> dict:
    """Launch the napari viewer to identify anatomical landmarks on a maximum intensity projection image.

    The following landmarks are identified for registration:
    - bregma: Just plain ol' bregma.
    - cFP: Frontal pole center.
    - rFP: Rightmost aspect of the frontal pole.
    - lFP: Leftmost aspect of the frontal pole.
    - rPB: Right lateral edge of the parietal bone.
    - lPB: Left lateral edge of the parietal bone.
    - lpRSP: Left posterior aspect of the retrosplenial cortex.
    - rpRSP: Right posterior aspect of the retrosplenial cortex.
    - aIPB: Anterior aspect of the interparietal bone.

    Args:
        maxip_image (npt.NDArray): Maximum intensity projection image. Could be either channel.
        alt_image (npt.NDArray): Alternative image to be displayed alongside the maximum intensity projection image. Usually a second channel.
        template_landmarks (dict, optional): Dictionary with the landmarks and their x-y coordinates. Dictionary keys are landmark names, while x-y coordinates are stored as an (y, x) tuple. Defaults to {}.

    Returns:
        dict: Dictionary with the landmarks and their x-y coordinates. Dictionary keys are landmark names, while x-y coordinates are stored as an (y, x) tuple.

    Raises:
        ValueError: If the maximum intensity projection and alternative image do not have the same dimensions.
    """
    if len(maxip_image.shape) == 3:
        maxip_height, maxip_width, _ = maxip_image.shape
    else:
        maxip_height, maxip_width = maxip_image.shape

    viewer = napari.view_image(maxip_image)

    if alt_image is not None:
        if len(alt_image.shape) == 3:
            alt_height, alt_width, _ = alt_image.shape
        else:
            alt_height, alt_width = alt_image.shape
        if (maxip_height, maxip_width) != (alt_height, alt_width):
            msg = "Maximum intensity projection and alternative image must have the same dimensions."
            raise ValueError(msg)

        viewer.add_image(alt_image, name="alt_maxip")

    default_landmark_locations = template_landmarks or {
        "bregma": (maxip_height / 2, maxip_width / 2),
        "cFP": (maxip_height / 7, maxip_width / 2),
        "rFP": (maxip_height / 7, maxip_width / 1.5),
        "lFP": (maxip_height / 7, maxip_width / 3),
        "rPB": (maxip_height / 4, maxip_width / 1.25),
        "lPB": (maxip_height / 4, maxip_width / 5),
        "lpRSP": (maxip_height / 1.25, maxip_width / 2.25),
        "rpRSP": (maxip_height / 1.25, maxip_width / 1.75),
        "aIPB": (maxip_height / 1.4, maxip_width / 2),
    }

    landmarks = list(default_landmark_locations.keys())

    points_layer = viewer.add_points(
        data=np.array(list(default_landmark_locations.values())),
        name="landmarks",
        ndim=2,
        properties={"label": landmarks},
        property_choices={"label": landmarks},
        symbol="o",
        face_color="label",
        face_color_cycle=COLOR_CYCLE,
        edge_width=0,  # fraction of point size
        size=5,
    )
    points_layer.face_color_mode = "cycle"

    # add the label menu widget to the viewer
    label_widget = _create_label_menu(points_layer, landmarks)
    viewer.window.add_dock_widget(label_widget)

    points_layer.mode = "select"

    napari.run()

    return OrderedDict(zip(landmarks, points_layer.data, strict=True))


def _create_label_menu(points_layer, labels):
    """Create a label menu widget that can be added to the napari viewer dock.

    Borrowed from https://napari.org/stable/tutorials/annotation/annotate_points.html.

    Args:
        points_layer (napari.layers.Points): A napari points layer.
        labels (List[str]): List of the labels for each keypoint to be annotated (e.g., the body parts to be labeled).

    Returns:
        Container: the magicgui Container with our dropdown menu widget
    """
    # Create the label selection menu
    label_menu = mgw.ComboBox(label="feature_label", choices=labels)
    info_text = mgw.Label(value="When you're done annotating,\nclose the viewer window.")
    close_button = mgw.Button(label="Save and Close")
    label_widget = mgw.Container(widgets=[label_menu, info_text, close_button])

    def close_viewer(event) -> None:
        napari.current_viewer().close()

    close_button.changed.connect(close_viewer)

    def update_label_menu(event) -> None:
        """Update the label menu when the point selection changes."""
        new_label = str(points_layer.current_properties["label"][0])
        if new_label != label_menu.value:
            label_menu.value = new_label

    points_layer.events.current_properties.connect(update_label_menu)

    def label_changed(new_label) -> None:
        """Update the Points layer when the label menu selection changes."""
        current_properties = points_layer.current_properties
        current_properties["label"] = np.asarray([new_label])
        points_layer.current_properties = current_properties
        points_layer.refresh_colors()

    label_menu.changed.connect(label_changed)

    return label_widget



#  Copyright (c) 2024 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import os

import dask
import numpy as np
import numpy.typing as npt
import zarr
from dask import array as da

from mesoscopy import io


def bin_array(
    array: da.Array | npt.NDArray,
    bins: int,
    interim_dir: str = ".",
    session_id: str = "null",
) -> zarr.Array:
    """Bin a 3D image array across its x and y axes.

    The function bins the width and height of a 3D image array by a factor of `bins`. It does not bin the z-axis (time).

    Args:
        array (Dask or NumPy Array): Array to be binned.
        bins (int): Number of bins in x and y directions (i.e. width and height).
        interim_dir (str or PathLike object, optional): Directory to store interim binned array data.
            Defaults to current working directory (".").
        session_id (str, optional): Session identifier for interim path. Defaults to "null".

    Returns:
        zarr.core.Array: Binned array as a persistent Zarr array object.
    """
    binned_array = array.reshape(  # pyright: ignore[reportAttributeAccessIssue]
        array.shape[0],  # pyright: ignore[reportArgumentType]
        1,
        int(array.shape[1] / bins),
        int(array.shape[1] // (array.shape[1] / bins)),
        int(array.shape[2] / bins),
        int(array.shape[2] // (array.shape[2] / bins)),
    ).mean(axis=(-1, 1, 3), dtype=np.float32)
    interim_path = f"{interim_dir}{os.sep}{session_id}_binned"
    return io.store_interim(binned_array, interim_path)


def frame_statistics(array: da.Array | npt.NDArray) -> tuple[npt.NDArray, npt.NDArray]:
    """Calculate mean and standard deviation for each frame in a 3D image array.

    Args:
        array (Dask or NumPy Array): Imaging array to calculate statistics for.

    Returns:
        tuple[npt.NDArray, npt.NDArray]: Tuple containing two NumPy arrays: means and standard deviations for each frame.
    """
    if type(array) is np.ndarray:
        array = da.from_array(array, chunks=(100, array.shape[1], array.shape[2]))  # pyright: ignore[reportArgumentType]

    frame_means, frame_stds = dask.compute(  # pyright: ignore[reportPrivateImportUsage]
        array.mean(axis=(1, 2), dtype=np.float32),
        array.std(axis=(1, 2), dtype=np.float32),
    )

    return frame_means, frame_stds


def channel_separation_filters(
    frame_means: npt.NDArray,
    frame_stds: npt.NDArray,
    use_means: bool = False,
    flip_channels: bool = False,
) -> tuple[list, list]:
    """Generate filters for separating channels based on frame means and standard deviations.

    Args:
        frame_means (npt.NDArray): Array of mean values for each frame.
        frame_stds (npt.NDArray): Array of standard deviation values for each frame.
        use_means (bool, optional): Use means instead of standard deviations for filtering. Defaults to False.
        flip_channels (bool, optional): Flip the channels. Defaults to False.

    Returns:
        tuple[list, list]: Tuple of two lists, containing the frame indices for each channel.
    """
    threshold = frame_stds.mean()
    gcamp_filter = np.nonzero(frame_stds > threshold)[0].tolist()
    isosb_filter = np.nonzero(frame_stds < threshold)[0].tolist()

    if use_means:
        threshold = frame_means.mean()
        gcamp_filter = np.nonzero(frame_means > threshold)[0].tolist()
        isosb_filter = np.nonzero(frame_means < threshold)[0].tolist()

    if flip_channels:
        return isosb_filter, gcamp_filter

    return gcamp_filter, isosb_filter


def rolling_dff(
    array: da.Array | npt.NDArray,
    window_width: int = 750,
    channel_name: str = "null",
    interim_dir: str = ".",
    session_id: str = "null",
) -> zarr.Array:
    """Calculate dF/F using a rolling window.

    Args:
        array (Dask or NumPy Array): Array to be separated. If array is a multi-channel recording,
            it needs to be filtered before it's passed to this function.
        window_width (int, optional): Window width for dF/F calculation. Defaults to 750.
        channel_name (str, optional): Channel name for interim path. Defaults to "null".
        interim_dir (str, optional): Directory to store interim dF/F data. Defaults to current working directory (".").
        session_id (str, optional): Session identifier for interim path. Defaults to "null".

    Returns:
        zarr.core.Array: dF/F array as a persistent Zarr array object.

    Raises:
        ValueError: If the window width is greater than the number of frames in the array.
    """
    if type(array) is np.ndarray:
        array = da.from_array(array, chunks=(100, array.shape[1], array.shape[2]))  # pyright: ignore[reportArgumentType]

    if window_width > len(array):
        msg = "Window width must be less than the number of frames."
        raise ValueError(msg)

    # If window width is an odd number, add 1 to make it even to avoid broadcast errors
    if window_width % 2 != 0:
        window_width += 1

    cumsum_vec = da.cumsum(array, dtype=np.uint32, axis=0)

    interim_path = interim_dir + os.sep + session_id + "_" + channel_name + "_cumsum"
    cumsum_vec = io.store_interim(cumsum_vec, interim_path)

    f0 = da.true_divide(
        (cumsum_vec[window_width:] - cumsum_vec[:-window_width]),
        window_width,
        dtype=np.float32,
    )

    interim_path = interim_dir + os.sep + session_id + "_" + channel_name + "_f0"
    f0 = io.store_interim(f0, interim_path)

    f0_start = da.mean(f0[: window_width // 2], axis=0).compute()
    padding_start = da.zeros((window_width // 2, *array.shape[1:3])) + f0_start
    f0_end = da.mean(f0[-(window_width // 2) :], axis=0).compute()
    padding_end = da.zeros((window_width // 2, *array.shape[1:3])) + f0_end

    f0 = da.insert(f0, [0], padding_start, axis=0)
    f0 = da.insert(f0, [f0.shape[0]], padding_end, axis=0)

    interim_path = interim_dir + os.sep + session_id + "_" + channel_name + "_f0_appended"
    f0 = io.store_interim(f0, interim_path)

    dff = da.true_divide(da.subtract(array, f0), f0, dtype=np.float32)

    interim_path = interim_dir + os.sep + session_id + "_" + channel_name + "_dff"

    return io.store_interim(dff, interim_path)


def projections(
    array: da.Array | npt.NDArray,
) -> dict[str, npt.NDArray]:
    """Calculate mean, standard deviation and maximum intensity projection frames.

    Args:
        array (npt.NDArray): Frame array to calculate projections for.

    Returns:
        dict[str, npt.NDArray]: Dictionary containing mean, standard deviation and maximum intensity projection frames.
    """
    mean_frame, std_frame, maxip = dask.compute(  # pyright: ignore[reportPrivateImportUsage]
        array.mean(axis=0),
        array.std(axis=0),
        array.max(axis=0),
    )

    return {
        "mean": mean_frame,
        "std": std_frame,
        "maxip": maxip,
    }



from datetime import datetime
from importlib import resources
from uuid import uuid4

import dask.array as da
import numpy as np
import pytest
from dateutil.tz import tzlocal
from pynwb import NWBHDF5IO
from pynwb import NWBFile

import mesoscopy.resources
from mesoscopy import io


def test_read_h5(raw_h5):
    _ = io.read_h5(raw_h5)


def test_h5_write(tmp_path):
    path = tmp_path / "test.h5"
    data = {"dataset1": np.array([1, 2, 3]), "dataset2": np.array([[1, 2], [3, 4]])}
    io.write_h5(path, data)
    assert io.read_h5(path)["dataset1"][0] == 1
    assert io.read_h5(path)["dataset2"][0][0] == 1


def test_read_nwb(nwbfile):
    _ = io.read_nwb(nwbfile)


def test_read_nwb_return_io(nwbfile):
    _, _ = io.read_nwb(nwbfile, return_io=True)


def test_write_nwb(tmp_path):
    path = tmp_path / "test.nwb"
    session_start_time = datetime(2024, 1, 1, 14, 0, 0, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="Test file, not real data",
        identifier=str(uuid4()),
        session_start_time=session_start_time,
        session_id="session_1234",
    )
    io.write_nwb(path, nwbfile, mode="w")
    _ = io.read_nwb(path)


def test_write_nwb_io(tmp_path):
    path = tmp_path / "test.nwb"
    session_start_time = datetime(2024, 1, 1, 14, 0, 0, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="Test file, not real data",
        identifier=str(uuid4()),
        session_start_time=session_start_time,
        session_id="session_1234",
    )
    io.write_nwb(path, nwbfile, mode="w", io=NWBHDF5IO(path, mode="w"))
    _ = io.read_nwb(path)


def test_store_interim(tmp_path):
    path = tmp_path / "test.zarr"
    data = np.random.rand(300, 142, 142)
    io.store_interim(data, str(path))


def test_store_interim_dask(tmp_path):
    path = tmp_path / "test.zarr"
    dask_data = da.from_array(np.random.rand(300, 142, 142), chunks=(100, 142, 142))
    io.store_interim(dask_data, str(path))


def test_load_interim(tmp_path):
    path = tmp_path / "test.h5"
    data = np.random.rand(300, 142, 142)
    assert io.store_interim(data, str(path)).any()


def test_read_points_fiji():
    assert io.read_points(str(resources.files(mesoscopy.resources).joinpath("ccf_template_top_140x142.points")))


def test_read_points_csv():
    assert io.read_points(str(resources.files(mesoscopy.resources).joinpath("ccf_template_landmarks_140x142.csv")))


def test_read_points_unsupported():
    with pytest.raises(ValueError):
        io.read_points("unsupported.file")


def test_write_points_csv(tmp_path):
    path = tmp_path / "test_points"
    data = {"testArea": [0, 200], "anotherTestArea": [250, 20]}
    io.write_points(str(path), data)



#  Copyright (c) 2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import json
from pathlib import Path

import click
import numpy as np
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

import mesoscopy.preprocess.qa as preqa
import mesoscopy.register.qa as regqa
from mesoscopy import io

PREPROCESSING_REPORT_TEMPLATE = "preprocessing.html"
REGISTRATION_REPORT_TEMPLATE = "registration.html"

env = Environment(loader=PackageLoader("mesoscopy.report", "templates"), autoescape=select_autoescape())


@click.command(name="report")
@click.argument("path")
@click.option(
    "-o",
    "--out_dir",
    type=click.Path(dir_okay=True),
    default=".",
    help="Output directory for preprocessed recording.",
)
def report_cmd(path: str, out_dir: str) -> str:
    """Generate a report for a mesoscopy processing step.

    Args:
        path (str): Path to the input file, which should be either a preprocessed or registered recording.
        out_dir (str): Directory where the report will be saved. Defaults to the current directory

    Returns:
        str: Path to the generated report.

    Raises:
        ValueError: If the input file does not match expected patterns for preprocessing or registration.
    """
    if path.endswith("_preprocessed.h5"):
        return generate_preprocessing_report(path, out_dir=out_dir)
    elif path.endswith("_registered.h5"):
        return generate_registration_report(path, out_dir=out_dir)
    else:
        raise ValueError


def generate_preprocessing_report(path: str, out_dir: str = ".") -> str:
    """Generate a preprocessing report for a mesoscopy recording.

    Args:
        path (str): Path to the preprocessed recording file.
        out_dir (str): Directory where the report will be saved. Defaults to the current directory.

    Returns:
        str: Path to the generated report.
    """
    preproc = io.read_h5(path)
    session_id = path.split("/")[-1].split("_preprocessed")[0]

    template = env.get_template(PREPROCESSING_REPORT_TEMPLATE)

    template_identifiers = {
        "session_id": session_id,
        "animal_id": session_id.split("_")[0].replace("sub-", ""),
        "session_date": session_id.split("_date-")[-1].replace("_", ":"),
        "experiment_id": session_id.split("_exp-")[-1].split("_")[0],
        "duration": str(
            (np.datetime64(preproc.get("timestamps")[-1]) - np.datetime64(preproc.get("timestamps")[0])).astype(  # type: ignore[attr-defined]
                "timedelta64[m]"
            )
        ),
        "frame_num": len(preproc.get("F")),  # type: ignore[attr-defined]
        "filesize": preproc.id.get_filesize() / (1024 * 1024),
        "fig_integrity_timestamps": preqa.plot_timestamps(
            [np.datetime64(ts) for ts in preproc.get("timestamps")],  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_pre_timeseries_mean": preqa.plot_raw_timeseries(
            preproc.get("qa").get("frame_means_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_pre_hist_mean": preqa.plot_raw_histogram(
            preproc.get("qa").get("frame_means_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_pre_hist_std": preqa.plot_raw_histogram(
            preproc.get("qa").get("frame_stds_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_timeseries": preqa.plot_channels_timeseries(
            gcamp_channel=preproc.get("qa").get("gcamp_mean_timeseries"),  # type: ignore[attr-defined]
            isosb_channel=preproc.get("qa").get("isosb_mean_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_gcamp_maxip": preqa.plot_projection(
            preproc.get("qa").get("gcamp_maxip_projection"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_gcamp_stdp": preqa.plot_projection(
            preproc.get("qa").get("gcamp_std_projection"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_isosb_maxip": preqa.plot_projection(
            preproc.get("qa").get("isosb_maxip_projection"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_isosb_stdp": preqa.plot_projection(
            preproc.get("qa").get("isosb_std_projection"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_filter_idx": preqa.plot_frame_ids(
            gcamp_ids=preproc.get("qa").get("gcamp_filter"),  # type: ignore[attr-defined]
            isosb_ids=preproc.get("qa").get("isosb_filter"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_separation_post_filter_pie": preqa.plot_filter_pie(
            gcamp_ids=preproc.get("qa").get("gcamp_filter"),  # type: ignore[attr-defined]
            isosb_ids=preproc.get("qa").get("isosb_filter"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_channel_dff": preqa.plot_channels_dff(
            gcamp_channel=preproc.get("qa").get("gcamp_dff_timeseries"),  # type: ignore[attr-defined]
            isosb_channel=preproc.get("qa").get("isosb_dff_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_corrected_dff": preqa.plot_corrected_dff(
            preproc.get("qa").get("f_mean_timeseries"),  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_corrected_example": preqa.plot_frame(
            preproc.get("F")[100],  # type: ignore[attr-defined]
            as_html=True,
        ),
        "fig_corrected_maxip": preqa.plot_projection(preproc.get("qa").get("f_maxip"), as_html=True),  # type: ignore[attr-defined]
    }

    out_path = out_dir / Path(path.split("/")[-1].replace(".h5", "_report.html"))
    out_path.write_text(template.render(template_identifiers), encoding="utf-8")
    return str(out_path)


def generate_registration_report(path: str, out_dir=".") -> str:
    """Generate a registration report for a mesoscopy recording.

    Args:
        path (str): Path to the registered recording file.
        out_dir (str): Directory where the report will be saved. Defaults to the current directory.

    Returns:
        str: Path to the generated report.
    """
    registered = io.read_h5(path)
    session_id = path.split("/")[-1].split("_preprocessed_registered")[0]

    template = env.get_template(REGISTRATION_REPORT_TEMPLATE)

    template_identifiers = {
        "session_id": session_id,
        "animal_id": session_id.split("_")[0].replace("sub-", ""),
        "session_date": session_id.split("_date-")[-1].replace("_", ":"),
        "experiment_id": session_id.split("_exp-")[-1].split("_")[0],
        "duration": str(
            (np.datetime64(registered.get("timestamps")[-1]) - np.datetime64(registered.get("timestamps")[0])).astype(  # type: ignore[attr-defined]
                "timedelta64[m]"
            )
        ),
        "frame_num": len(registered.get("F")),  # type: ignore[attr-defined]
        "filesize": registered.id.get_filesize() / (1024 * 1024),
        "fig_landmarks_unregistered": regqa.plot_landmarks(
            source_landmarks=registered.get("/qa/template_landmarks"),
            target_landmarks=registered.get("/qa/recording_landmarks"),
            as_html=True,
        ),
        "fig_landmarks_registered": regqa.plot_landmarks(
            source_landmarks=registered.get("/qa/template_landmarks"),
            target_landmarks=registered.get("/qa/registered_landmarks"),
            as_html=True,
        ),
        "fig_frame_example": regqa.plot_frame(
            registered.get("F")[1], landmarks=registered.get("/qa/template_landmarks"), as_html=True
        ),
    }

    out_path = out_dir / Path(path.split("/")[-1].replace(".h5", "_report.html"))
    out_path.write_text(template.render(template_identifiers), encoding="utf-8")
    return str(out_path)



#  Copyright (c) 2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import numpy as np
import scipy.ndimage as ndi


def laplace_gaussian(
    deltaf_series: np.ndarray,
    sigma: int = 2,
    **kwargs,
):
    """Spatial smoothing using a Laplace of Gaussian filter.

    Args:
        deltaf_series (np.ndarray): DeltaF/F series.
        sigma (int, optional): Standard deviation for Gaussian kernel. Defaults to 2.
        **kwargs: Additional keyword arguments for `scipy.ndimage.gaussian_laplace`.

    Returns:
        np.ndarray: Smoothed DeltaF/F series.
    """
    sigma_iso = (0, sigma, sigma)

    return ndi.gaussian_laplace(deltaf_series, sigma=sigma_iso, **kwargs)



#  Copyright (c) 2022 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import json
import os
import pathlib

import click
import h5py
import numpy as np
from pynwb import TimeSeries
from pynwb.image import ImageSeries
from pynwb.ophys import CorrectedImageStack

import mesoscopy.preprocess as preproc
import mesoscopy.preprocess.compute as preproc_compute
import mesoscopy.register.landmarks_gui as reg_gui
import mesoscopy.register.transform as trf
import mesoscopy.resources as res
from mesoscopy import io
from mesoscopy import timer as timer


@click.group("register")
def register_cmd() -> None:
    """Register recordings to a template."""


@register_cmd.command("label")
@click.argument(
    "path",
    type=click.Path(exists=True),
)
@click.option(
    "-o",
    "--out_dir",
    type=click.Path(dir_okay=True),
    default="./",
    help="Output directory for registered recording.",
)
@click.option(
    "-t",
    "--template-points",
    type=click.Path(dir_okay=False),
    help="Path to template landmark points in CSV or Fiji XML points format",
)
@click.option(
    "--session-id",
    type=str,
    help="Session ID for the recording.",
)
def label_cmd(path, out_dir, template_points, session_id) -> dict:
    """Mark landmarks on a recording for registration to a template using the landmarks GUI.

    Args:
        path (str): Path to preprocessed HDF5 file or NWB file.
        out_dir (str): Output directory for registration landmarks file.
        template_points (str): Path to template landmark points in CSV or Fiji XML points format.
        session_id (str): Session ID for the recording.

    Returns:
        dict: Dictionary with the landmarks and their x-y coordinates.
              Dictionary keys are landmark names, while x-y coordinates are stored as an (y, x)
    """
    click.echo("Loading imaging data...")
    nwb = bool(path.endswith(".nwb"))

    if not session_id:
        session_id = path.split("/")[-1].replace(".nwb", "") if nwb else path.split("/")[-1].replace(".h5", "")
        session_id = session_id.replace("_preprocessed", "")

    maxip = None
    isosb_maxip = None

    # Load maxip from preprocessed file.
    # if it does not exist, generate maxip from raw data.
    if path.endswith("_preprocessed.h5"):
        maxip, isosb_maxip = load_maxips(path)
    else:
        # generate maxip from raw data
        click.echo("⚠️ No maximum intensity projection found. Generating from raw data, this might take some time...")
        with timer.Timer("Generating maximum intensity projection"):
            _, raw_data, _ = preproc.load_raw(path, nwb=nwb)
            maxip = preproc_compute.projections(raw_data)["maxip"]

    click.echo("Loading template landmarks...")
    template_landmarks = res.get_default_landmarks()
    if template_points:
        template_landmarks = io.read_points(template_points)

    click.echo("Launching landmark identification GUI...")
    recording_landmarks = reg_gui.mark_landmarks(maxip, isosb_maxip, template_landmarks)

    click.echo("Saving recording landmarks...")
    outpath = out_dir + os.sep + session_id + "_landmarks.csv"
    io.write_points(outpath, recording_landmarks)

    click.echo(f"Recording landmarks saved at {outpath}.")

    return recording_landmarks


@register_cmd.command("landmarks")
@click.argument(
    "path",
    type=click.Path(exists=True),
)
@click.option(
    "-o",
    "--out_dir",
    type=click.Path(dir_okay=True),
    default="./",
    help="Output directory for registered recording.",
)
@click.option(
    "-r",
    "--recording-points",
    type=click.Path(dir_okay=False),
    help="Path to recording landmark points in Fiji XML points format",
)
@click.option(
    "-t",
    "--template-points",
    type=click.Path(dir_okay=False),
    help="Path to template landmark points in Fiji XML points format",
)
@click.option("--crop-x", default=0, help="Crop recording along the x-axis.")
@click.option("--crop-y", default=0, help="Crop recording along the y-axis.")
def landmarks_cmd(
    path: str,
    out_dir: str,
    recording_points: str,
    template_points: str,
    crop_x: int = 0,
    crop_y: int = 0,
) -> str:
    """Register a recording to a template based on defined landmarks.

    Args:
        path (str): Path to preprocessed recording HDF5 or NWB file.
        out_dir (str): Output directory for registered recording.
        recording_points (str, optional): Path to recording landmark points in CSV or Fiji XML points format.
        template_points (str, optional): Path to template landmark points in CSV or Fiji XML points format.
        crop_x (int, optional): Number of pixels to crop from the x-axis of the recording. Defaults to 0.
        crop_y (int, optional): Number of pixels to crop from the y-axis of the recording. Defaults to 0.

    Returns:
        str: Path to the registered recording file.

    Raises:
        ValueError: If the path to recording landmarks cannot be inferred.
    """
    click.echo(f"Registering recording {path} to template.")

    os.makedirs(out_dir, exist_ok=True)

    click.echo("Loading imaging data...")

    # Determine whether we're working with an NWB file
    nwb = True if path.endswith(".nwb") else False

    session_id, deltaf_series, timestamps = load_deltaf(path, nwb)

    click.echo("Loading landmarks...")
    template_landmarks = res.get_default_landmarks()
    if template_points:
        template_landmarks = io.read_points(template_points)

    if not recording_points:
        if nwb and pathlib.Path(path.replace(".nwb", "_landmarks.csv")).exists():
            recording_points = path.replace(".nwb", "_landmarks.csv")
        elif pathlib.Path(path.replace(".h5", "_landmarks.csv")).exists():
            recording_points = path.replace(".h5", "_landmarks.csv")
        else:
            msg = "Path to recording landmarks could not be inferred. Please supply a recording landmarks file."
            raise ValueError(msg)
    recording_landmarks = io.read_points(recording_points)

    warped, tform = trf.landmarks_affine(
        deltaf_series,
        recording_landmarks,
        template_landmarks,
        crop_x=crop_x,
        crop_y=crop_y,
    )

    # Save warped frames and timestamps
    outpath = out_dir + os.sep + session_id + "_registered.h5"
    outpath = io.write_h5(
        path=outpath,
        data={
            "/F": warped,
            "/timestamps": timestamps,
            "/tform": tform.params,
            "/qa/recording_landmarks": np.array(list(recording_landmarks.values())),
            "/qa/template_landmarks": np.array(list(template_landmarks.values())),
            "/qa/registered_landmarks": tform.inverse(np.array(list(recording_landmarks.values()))),
        },
    )
    click.echo(f"Saved registered frames at {outpath}")

    if nwb:
        click.echo("Updating NWB file...")
        update_nwb(path, outpath, tform)
        click.echo(f"Updated NWB file at {path}")

    return outpath


def load_deltaf(path: str, nwb: bool = False) -> tuple[str, np.ndarray, np.ndarray]:
    """Load preprocessed deltaf from an HDF5 or NWB file.

    Args:
        path (str): Path to the preprocessed file.
        nwb (bool, optional): Whether the file is an NWB file. Defaults to False.

    Returns:
        tuple[str, np.ndarray, np.ndarray]: Session identifier, dF/F series, and timestamps.
    """
    if nwb:
        nwbfile = io.read_nwb(path)
        session_id = nwbfile.identifier
        deltaf_series = nwbfile.processing["ophys"]["DeltaFSeries"].data
        timestamps = nwbfile.processing["ophys"]["DeltaFSeries"].timestamps
    else:
        session_id = path.split("/")[-1].replace(".h5", "")
        f_preproc = h5py.File(path)
        deltaf_series = f_preproc["/F"]
        timestamps = f_preproc["/timestamps"]

    return session_id, deltaf_series, timestamps


def load_maxips(path: str) -> tuple[np.ndarray, np.ndarray]:
    """Load maximum intensity projections from a preprocessed HDF5 file.

    Args:
        path (str): Path to the preprocessed HDF5 file.

    Returns:
        tuple[np.ndarray, np.ndarray]: Maximum intensity projection for gcamp and isosb channels.
    """
    f_preproc = h5py.File(path, "r")
    gcamp_maxip_projection = np.array(f_preproc["/qa/gcamp_maxip_projection"])
    isosb_maxip_projection = np.array(f_preproc["/qa/isosb_maxip_projection"])

    return gcamp_maxip_projection, isosb_maxip_projection


def update_nwb(nwb_path: str, h5_path: str, tform_params: np.ndarray) -> None:
    """Update an NWB file with registered imaging data stored in an HDF5 file.

    Creates a link between the NWB file and the HDF5 file. See https://pynwb.readthedocs.io/en/stable/tutorials/advanced_io/linking_data.html.

    Args:
        nwb_path (str): Path to the NWB file.
        h5_path (str): Path to the HDF5 file containing the registered images.
        tform_params (np.ndarray): Affine transformation parameters.
    """
    nwbfile, nwbio = io.read_nwb(nwb_path, return_io=True)
    f = h5py.File(h5_path, "r")

    try:
        ophys_module = nwbfile.create_processing_module(name="ophys", description="optical physiology processed data")
    except ValueError:
        click.echo("Processing module already exists...")
        ophys_module = nwbfile.processing["ophys"]

    registered_series = ImageSeries(
        name="corrected",
        data=f["/F"],
        timestamps=f["/timestamps"],
        unit="df/f",
        description="dF/F widefield cortical imaging series.",
        comments="This is the haemodynamic corrected series registered to the Allen Brain Atlas CCFv3.",
    )

    xy_translation = TimeSeries(
        name="xy_translation",
        data=np.repeat(tform_params, len(f["/timestamps"]), axis=0),
        unit="pixels",
        timestamps=f["/timestamps"],
        description="Affine transformation parameters for image registration to the ABA CCFv3.",
    )

    corrected_image_stack = CorrectedImageStack(
        name="CCFRegisteredSeries",
        corrected=registered_series,
        original=nwbfile.acquisition["DualChannelImagingSeries"],
        xy_translation=xy_translation,
    )

    ophys_module.add(corrected_image_stack)

    io.write_nwb(nwb_path, nwbfile, io=nwbio)

    return nwbfile



#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""Animal metadata readers."""

import json

import yaml

DEFAULT_METADATA = {
    "subject_id": "unknown",
    "sex": "Unknown",
    "genotype": "Unknown",
    "species": "Unknown",
    "strain": "Unknown",
    "dob": "1900-01-01",
    "session_description": "No description",
    "experimenter": "Unknown",
    "lab": "Unknown",
    "institution": "Unknown",
}


def read_yaml(path: str) -> dict:
    """Read a yaml metadata file.

    Args:
        path (str): Path to file.

    Raises:
        ValueError: Raised if file contains fields not in DEFAULT_METADATA.

    Returns:
        dict: Subject metadata.
    """
    with open(path) as fp:
        metadata = yaml.safe_load(fp)

        for key in metadata.keys():
            if key not in list(DEFAULT_METADATA.keys()):
                raise ValueError(f"Invalid file contents - unrecognised key <{key}>.")

        for field in DEFAULT_METADATA:
            if field not in metadata.keys():
                metadata[field] = DEFAULT_METADATA.get(field)

        return metadata


def read_json(path: str) -> dict:
    """Read a json metadata file.

    Args:
        path (str): Path to file.

    Raises:
        ValueError: Raised if file contains fields not in DEFAULT_METADATA.

    Returns:
        dict: Subject metadata.
    """
    with open(path) as fp:
        metadata = json.load(fp)

        for key in metadata.keys():
            if key not in list(DEFAULT_METADATA.keys()):
                raise ValueError(f"Invalid file contents - unrecognised key <{key}>.")

        for field in DEFAULT_METADATA:
            if field not in metadata.keys():
                metadata[field] = DEFAULT_METADATA.get(field)

        return metadata



#  Copyright (c) 2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import time

import click


class Timer:
    """Timer context manager for measuring execution time of analysis blocks."""

    def __init__(self, message: str = "Analysis"):
        self.message = message

    def __enter__(self):
        self.start_time = time.time()
        click.echo(f"{self.message}...")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        elapsed_time = time.time() - self.start_time
        if elapsed_time > 60:
            minutes, seconds = divmod(elapsed_time, 60)
            click.echo(f"{self.message} took {int(minutes)} minutes and {seconds:.2f} seconds.")
        else:
            click.echo(f"{self.message} took {elapsed_time:.2f} seconds.")


class TimerError(Exception):
    """Custom exception for timer errors."""



from click.testing import CliRunner

import mesoscopy
import mesoscopy.inspect as insp


def test_inspect_meta(preproc_nwb):
    insp.inspect(input_path=preproc_nwb, info_level="meta_only")


def test_inspect_meta_noprocessing(nwbfile):
    insp.inspect(input_path=nwbfile, info_level="meta_only")


# def test_inspect_acquisition(nwbfile):
#     insp.inspect(input_path=nwbfile, info_level="acquisition")


# def test_inspect_acquisition_notfound(nwbfile_noacquisition):
#     insp.inspect(input_path=nwbfile_noacquisition, info_level="acquisition")


# def test_inspect_deltaf(preproc_nwb):
#     insp.inspect(input_path=preproc_nwb, info_level="deltaf")


# def test_inspect_deltaf_notfound(nwbfile):
#     insp.inspect(input_path=nwbfile, info_level="deltaf")


# def test_inspect_registration(): ...


# def test_inspect_registration_notfound(nwbfile): ...


def test_cmd_parity(nwbfile):
    """Check if click preprocess cmd caller works"""
    runner = CliRunner()
    result = runner.invoke(
        mesoscopy.cli,
        args=f"inspect {nwbfile}",
    )
    assert result.exit_code == 0



#  Copyright (c) 2022 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#  following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

"""Processing submodule."""

import os
import click
import numpy as np
import dask.array as da

import mesoscopy.timer as timer

import mesoscopy.io as io
import mesoscopy.process.smooth as psm


@click.group("process")
def process_cmd(): ...


@process_cmd.command("smooth")
@click.argument(
    "path",
    type=click.Path(exists=True),
)
@click.option(
    "-o",
    "--out_dir",
    type=click.Path(dir_okay=True),
    default="./",
    help="Output directory for smoothed recording.",
)
@click.option(
    "-s",
    "--sigma",
    type=int,
    default=2,
    help="Output directory for smoothed recording.",
)
def smooth_cmd(path: str, out_dir: str, sigma: int = 2):
    """Generate a smoothed DeltaF/F recording using a Laplace of Gaussian filter."""
    if not os.path.exists(out_dir):
        click.echo(f"Creating output directory {out_dir}...")
        os.makedirs(out_dir)

    click.echo(f"Loading preprocessed recording from {path}...")
    # Determine whether we're working with an NWB file
    nwb = bool(path.endswith(".nwb"))
    session_id, deltaf_series, timestamps = load_deltaf(path, nwb=nwb)

    outpath = out_dir + os.sep + session_id + "_smoothed.h5"

    with timer.Timer(message="Smoothing with LoG"):
        smoothed_deltaf = psm.laplace_gaussian(deltaf_series)

        outpath = io.write_h5(
            path=outpath,
            data={
                "/F": smoothed_deltaf,
                "/timestamps": timestamps,
            },
        )
    click.echo(f"Saved smoothed recording at {outpath}")

    # if nwb:
    #     click.echo("Updating NWB file...")
    #     update_nwb(path, outpath, tform)
    #     click.echo(f"Updated NWB file at {path}")

    return outpath


def load_deltaf(path: str, nwb: bool = False) -> tuple[str, np.ndarray, np.ndarray]:
    """Load preprocessed deltaf from an HDF5 or NWB file.

    Args:
        path (str): Path to the preprocessed file.
        nwb (bool, optional): Whether the file is an NWB file. Defaults to False.

    Returns:
        tuple[str, np.ndarray, np.ndarray]: Session identifier, dF/F series, and timestamps.
    """
    if nwb:
        nwbfile = io.read_nwb(path)
        session_id = nwbfile.identifier
        deltaf_series = nwbfile.processing["ophys"]["DeltaFSeries"].data
        timestamps = nwbfile.processing["ophys"]["DeltaFSeries"].timestamps
    else:
        session_id = path.split("/")[-1].replace(".h5", "")
        f_preproc = io.read_h5(path)
        deltaf_series = f_preproc["/F"]
        timestamps = f_preproc["/timestamps"]

    return session_id, np.array(deltaf_series), np.array(timestamps)



import os
import pathlib

from click.testing import CliRunner
from pynwb import NWBHDF5IO

import mesoscopy
import mesoscopy.convert.hdf5 as conv_h5
import mesoscopy.convert.metadata as mtd
import mesoscopy.convert.video as conv_vid


def test_metadata_yaml(meta_yaml):
    expected_keys = list(mtd.DEFAULT_METADATA.keys())
    meta = mtd.read_yaml(meta_yaml)
    assert sorted(expected_keys) == sorted(list(meta.keys()))
    assert meta.get("subject_id") == "testyaml"
    assert meta.get("experimenter") == "John Doe"


def test_metadata_json(meta_json):
    expected_keys = list(mtd.DEFAULT_METADATA.keys())
    meta = mtd.read_json(meta_json)
    assert sorted(expected_keys) == sorted(list(meta.keys()))
    assert meta.get("subject_id") == "testjson"
    assert meta.get("experimenter") == "John Doe"


def test_convert_h5_linkonly(raw_h5, output_dir):
    expected_outpath = output_dir + os.sep + raw_h5.split(os.sep)[-1].replace(".h5", ".nwb")
    nwb_outpath = conv_h5.to_nwb(raw_h5, out_dir=output_dir, link_only=True)
    assert expected_outpath == nwb_outpath
    assert pathlib.Path(nwb_outpath).is_file()
    assert pathlib.Path(raw_h5).stat().st_size > pathlib.Path(nwb_outpath).stat().st_size


def test_convert_h5_eagercopy(raw_h5, output_dir):
    expected_outpath = output_dir + os.sep + raw_h5.split(os.sep)[-1].replace(".h5", ".nwb")
    nwb_outpath = conv_h5.to_nwb(raw_h5, out_dir=output_dir)
    assert expected_outpath == nwb_outpath
    assert pathlib.Path(nwb_outpath).is_file()
    assert pathlib.Path(raw_h5).stat().st_size <= pathlib.Path(nwb_outpath).stat().st_size


def test_convert_h5_metadata_yaml(raw_h5, output_dir, meta_yaml):
    nwb_outpath = conv_h5.to_nwb(raw_h5, out_dir=output_dir, meta_path=meta_yaml)

    io = NWBHDF5IO(nwb_outpath, mode="r")
    nwbfile = io.read()
    assert nwbfile.subject.subject_id == "testyaml"


def test_convert_h5_metadata_json(raw_h5, output_dir, meta_json):
    nwb_outpath = conv_h5.to_nwb(raw_h5, out_dir=output_dir, meta_path=meta_json)

    io = NWBHDF5IO(nwb_outpath, mode="r")
    nwbfile = io.read()
    assert nwbfile.subject.subject_id == "testjson"


def test_convert_h5_metadata_args(raw_h5, output_dir):
    nwb_outpath = conv_h5.to_nwb(
        raw_h5,
        out_dir=output_dir,
        subject_id="sometest",
        sex="F",
        genotype="Wt",
        species="Mus",
        strain="c57",
        dob="1999-01-02",
        session_description="somedesc",
        experimenter="Python Test",
        lab="well funded lab",
        institution="University of Great Science",
    )

    io = NWBHDF5IO(nwb_outpath, mode="r")
    nwbfile = io.read()
    assert nwbfile.subject.subject_id == "sometest"
    assert nwbfile.session_description == "somedesc"
    assert nwbfile.experimenter == ("Python Test",)


def test_convert_h5_metadata_args_file_mixed(raw_h5, output_dir): ...


def test_h5_cmd_parity(raw_h5, output_dir):
    """Check if click convert hdf5 cmd caller works"""
    runner = CliRunner()
    result = runner.invoke(
        mesoscopy.cli,
        args=f"convert h5 --out-dir {output_dir} {raw_h5}",
    )
    assert result.exit_code == 0


def test_avi_h5_converter(raw_avi, output_dir):
    h5_outpath = conv_vid.to_hdf5(raw_avi, out_dir=output_dir)
    assert pathlib.Path(h5_outpath).is_file()


def test_avi_h5_converter_with_timestamps(raw_avi, raw_timestamps, output_dir):
    h5_outpath = conv_vid.to_hdf5(raw_avi, output_dir, ts_path=raw_timestamps)
    assert pathlib.Path(h5_outpath).is_file()


def test_avi_nwb_converter(raw_avi, output_dir):
    h5_outpath = conv_vid.to_nwb(raw_avi, out_dir=output_dir)
    assert pathlib.Path(h5_outpath).is_file()



#  Copyright (c) 2022-2025 Constantinos Eleftheriou <Constantinos.Eleftheriou@ed.ac.uk>.
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy of this
#   software and associated documentation files (the "Software"), to deal in the
#   Software without restriction, including without limitation the rights to use, copy,
#   modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#   and to permit persons to whom the Software is furnished to do so, subject to the
#   following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies
#  or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
#  IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from pynwb import NWBHDF5IO


def export_standalone(nwb_path: str, out_path: str = "") -> str:
    """Create a sharable copy of an NWB file by resolving external data links.

    This function reads an NWB file, generates a new unique identifier for the NWBFile,
    and exports it to a new file with all external data links resolved (i.e., all data
    is contained within the new file). This is useful for sharing NWB files without
    dependencies on external files (e.g. for uploading on DANDI).

    Args:
        nwb_path (str): Path to the source NWB file.
        out_path (str, optional): Path to save the exported NWB file. If not provided,
            the output file will be named as the input file with '_export.nwb' appended.

    Returns:
        str: Path to the exported NWB file.
    """
    with NWBHDF5IO(nwb_path, mode="r") as read_io:
        nwbfile = read_io.read()
        nwbfile.generate_new_id()

        if not out_path:
            out_path = nwb_path.replace(".nwb", "_export.nwb")
        with NWBHDF5IO(out_path, mode="w") as export_io:
            export_io.export(src_io=read_io, nwbfile=nwbfile, write_args={"link_data": False})

    return out_path



import os
import pathlib

import pytest
from click.testing import CliRunner

import mesoscopy
import mesoscopy.export.nwb as nwb_export


def test_nwb_sharing_export(preproc_nwb, output_dir):
    """Test the nwb_sharing_export function."""
    out_path = os.path.join(output_dir, "test_export.nwb")
    exported_path = nwb_export.export_standalone(preproc_nwb, out_path)
    assert exported_path.endswith("_export.nwb")
    assert exported_path != preproc_nwb

    # Clean up the created file
    if pathlib.Path(exported_path).exists():
        pathlib.Path(exported_path).unlink()


def test_nwb_sharing_export_no_outpath(preproc_nwb):
    """Test the nwb_sharing_export function without specifying an output path."""
    exported_path = nwb_export.export_standalone(preproc_nwb)
    assert exported_path.endswith("_export.nwb")
    assert exported_path != preproc_nwb

    # Clean up the created file
    if pathlib.Path(exported_path).exists():
        pathlib.Path(exported_path).unlink()


def test_nwb_sharing_export_invalid_path():
    """Test the nwb_sharing_export function with an invalid NWB file path."""
    with pytest.raises(FileNotFoundError):
        nwb_export.export_standalone("invalid_path.nwb")

    # Ensure no file is created
    assert not pathlib.Path("invalid_path_export.nwb").exists()


def test_nwb_sharing_cmd_parity(preproc_nwb, output_dir):
    """Check if click export nwb cmd caller works."""
    runner = CliRunner()
    out_path = os.path.join(output_dir, "test_export.nwb")
    result = runner.invoke(
        mesoscopy.cli,
        args=f"export nwb {preproc_nwb} --out-path {out_path}",
    )
    assert result.exit_code == 0
    assert "Exporting NWB file..." in result.output
    assert "NWB file exported to" in result.output

    exported_path = os.path.join(output_dir, "test_export.nwb")
    assert pathlib.Path(exported_path).is_file()

    # Clean up the created file
    if pathlib.Path(exported_path).exists():
        pathlib.Path(exported_path).unlink()
